local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia", true)
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 95 $")

local d = Utopia.debugprint
local module = Utopia.modules.UpTime
local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour

local bgDef = {edgeInsets = {left = 6, right = 6, top = 6, bottom = 6}, edgeSize = 10, tileSize = 32, edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border"}
function module:SetBackdrop(frame, back)
	bgDef.bgFile = back and "Interface\\Tooltips\\UI-Tooltip-Background" or nil
	frame:SetBackdrop(bgDef)
	frame:SetBackdropBorderColor(1, 1, 1, 0.8)
	if (back) then
		frame:SetBackdropColor(0, 0, 0, 1)
	end
end	

do
	local function frameDragStop(self)
		self:StopMovingOrSizing()
		Utopia:SavePosition(self)
	end

	local function frameGripDragStart(self)
		self:GetParent():StartSizing()
	end

	local function frameGripDragStop(self)
		self:GetParent():StopMovingOrSizing()
		Utopia:SavePosition(self:GetParent())
	end

	-- MovableFrame
	function module:MovableFrame(frame)
		tinsert(UISpecialFrames, frame:GetName())

		frame:SetToplevel(true)
		frame:SetMovable(true)
		frame:SetResizable(true)
		frame:EnableMouse(true)
		frame:RegisterForDrag("LeftButton")

		frame.grip = CreateFrame("Frame", nil, frame)
		frame.grip:SetPoint("BOTTOMRIGHT", -4, 7)
		frame.grip:SetWidth(20)
		frame.grip:SetHeight(20)
		frame.grip.tex = frame.grip:CreateTexture(nil, "OVERLAY")
		frame.grip.tex:SetAllPoints()
		frame.grip.tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\ResizeGrip")
		frame.grip.tex:SetBlendMode("ADD")
		frame.grip.hl = frame.grip:CreateTexture(nil, "HIGHLIGHT")
		frame.grip.hl:SetAllPoints()
		frame.grip.hl:SetTexture("Interface\\Addons\\Utopia\\Textures\\ResizeGrip")
		frame.grip.hl:SetBlendMode("ADD")

		frame:SetScript("OnDragStart", frame.StartMoving)
		frame:SetScript("OnDragStop", frameDragStop)

		frame.grip:EnableMouse(true)
		frame.grip:RegisterForDrag("LeftButton")
		frame.grip:SetScript("OnDragStart", frameGripDragStart)
		frame.grip:SetScript("OnDragStop", frameGripDragStop)

		Utopia:RestorePosition(frame)
	end
end

-- CreateArrowButton
function module:CreateArrowButton(parent, which)
	local template
	if (which == "up") then
		template = "UIPanelScrollUpButtonTemplate"
	elseif (which == "down") then
		template = "UIPanelScrollDownButtonTemplate"
	end

	local b = CreateFrame("Button", nil, parent, template)
	return b
end

local function genericOnLeave(self)
	GameTooltip:Hide()
end

do
	local function sliderOnEnter(self)
		self.text:SetTextColor(1, 1, 1)
		if (self.tooltip) then
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
			GameTooltip:SetText(self.text:GetText(), 1, 1, 1)
			GameTooltip:AddLine(self.tooltip, nil, nil, nil, 1)
			GameTooltip:Show()
		end
	end

	local function sliderOnLeave(self)
		self.text:SetTextColor(1, 0.82, 0)
		GameTooltip:Hide()
	end

	local function sliderOnValueChanged(self, value)
		if (self.OnValueChanged) then
			self.OnValueChanged(self:GetParent(), value)
		end
		self.midText:SetText(value)
	end
	
	function module:Slider(parent, text, get, set, Min, Max, step, anchor, x, y, relFrame, rel)
--@debug@
		assert(type(parent) == "table" and parent.GetFrameType)
		assert(type(text) == "string")
		assert(type(get) == "function")
		assert(type(set) == "function")
		assert(type(Min) == "number")
		assert(type(Max) == "number")
		assert(type(step) == "number")
		assert(type(anchor) == "string")
		assert(not x or type(x) == "number")
		assert(not y or type(y) == "number")
		assert(not relFrame or (type(relFrame) == "table" and relFrame.GetFrameType))
		assert(not rel or type(rel) == "string")
--@debug@

		local b = CreateFrame("Slider", nil, parent, "OptionsSliderTemplate")
		b:SetWidth(130)

		local temp = new(b:GetRegions())
		for i,reg in ipairs(temp) do
			if (reg:GetObjectType() == "FontString") then
				local point, parent, relpoint, x, y = reg:GetPoint(1)
				if (point == "BOTTOM" and relpoint == "TOP") then
					b.text = reg
				elseif (point == "TOPLEFT" and relpoint == "BOTTOMLEFT") then
					b.leftText = reg
				elseif (point == "TOPRIGHT" and relpoint == "BOTTOMRIGHT") then
					b.rightText = reg
				end
			end
		end

		b.text:SetTextColor(1, 0.82, 0)

		b.midText = b:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		b.midText:SetPoint("TOPLEFT", b.leftText, "TOPRIGHT")
		b.midText:SetPoint("BOTTOMRIGHT", b.rightText, "BOTTOMLEFT")
		b.midText:SetTextColor(0.7, 0.7, 0.7)

		b:SetScript("OnEnter", sliderOnEnter)
		b:SetScript("OnLeave", sliderOnLeave)
		b:SetScript("OnValueChanged", sliderOnValueChanged)

		b.text:SetText(text)
		b.leftText:SetText(Min)
		b.rightText:SetText(Max)
		b:SetPoint(anchor, relFrame or parent, rel or anchor, x, y)

		b:SetMinMaxValues(Min, Max)
		b:SetValueStep(step)
		b:SetValue(get(parent))
		
		b.OnValueChanged = set

		return b
	end
end

do
	local function buttonOnEnter(self)
		if (self.tooltip) then
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
			GameTooltip:SetText(self:GetText(), 1, 1, 1)
			GameTooltip:AddLine(self.tooltip, nil, nil, nil, 1)
			GameTooltip:Show()
		end
	end

	-- Button
	function module:Button(parent, text, click, anchor, x, y, relFrame, rel)
--@debug@
		assert(type(parent) == "table" and parent.GetFrameType)
		assert(type(text) == "string")
		assert(type(click) == "function")
		assert(type(anchor) == "string")
		assert(not x or type(x) == "number")
		assert(not y or type(y) == "number")
		assert(not relFrame or (type(relFrame) == "table" and relFrame.GetFrameType))
		assert(not rel or type(rel) == "string")
--@debug@

		local b = CreateFrame("Button", nil, parent, "OptionsButtonTemplate")
		b.text = b:GetRegions()
		b.text:SetAllPoints(b)			-- Makes the text part (first region) fit all over button, instead of just centered and fuxed when scaled
		b:SetScript("OnClick", click)
		b:SetScript("OnEnter", buttonOnEnter)
		b:SetScript("OnLeave", genericOnLeave)
		b:SetText(text)
		b:SetWidth(max(100, b.text:GetStringWidth() + 25))
		b:SetHeight(26)
		b:SetPoint(anchor, relFrame or parent, rel or anchor, x, y)
		return b
	end

	-- checkboxOnClick
	local function checkboxOnClick(self)
		self.set(self, self:GetChecked())
	end

	-- checkboxOnShow
	local function checkboxOnShow(self)
		self:SetChecked(self.get(self))
	end

	local function checkboxOnEnter(self)
		self.text:SetTextColor(1, 1, 1)
		if (self.tooltip) then
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
			GameTooltip:SetText(self.text:GetText(), 1, 1, 1)
			GameTooltip:AddLine(self.tooltip, nil, nil, nil, 1)
			GameTooltip:Show()
		end
	end

	local function checkboxOnLeave(self)
		self.text:SetTextColor(1, 0.82, 0)
		GameTooltip:Hide()
	end

	local function checkboxSetEnabled(self, enabled)
		if (enabled) then
			self:Enable()
			self.text:SetTextColor(1, 0.82, 0)
		else
			self:Disable()
			self.text:SetTextColor(0.5, 0.5, 0.5)
		end
	end

	-- CheckBox
	function module:CheckBox(parent, text, get, set, anchor, x, y, relFrame, rel)
--@debug@
		assert(type(parent) == "table" and parent.GetFrameType)
		assert(type(text) == "string")
		assert(type(get) == "function")
		assert(type(set) == "function")
		assert(type(anchor) == "string")
		assert(not x or type(x) == "number")
		assert(not y or type(y) == "number")
		assert(not relFrame or (type(relFrame) == "table" and relFrame.GetFrameType))
		assert(not rel or type(rel) == "string")
--@debug@

		local b = CreateFrame("CheckButton", nil, parent, "OptionsCheckButtonTemplate")
		local temp = new(b:GetRegions())
		for i,reg in ipairs(temp) do
			if (reg:GetObjectType() == "FontString") then
				b.text = reg
				break
			end
		end
		assert(b.text)
		b:SetScript("OnClick", checkboxOnClick)
		b:SetScript("OnShow", checkboxOnShow)
		b:SetScript("OnEnter", checkboxOnEnter)
		b:SetScript("OnLeave", checkboxOnLeave)
		b.SetEnabled = checkboxSetEnabled

		b.set = set
		b.get = get
		b.text:SetText(text)
		b:SetHeight(26)
		b:SetWidth(26)
		b:SetPoint(anchor, relFrame or parent, rel or anchor, x, y)
		b:SetHitRectInsets(0, -(b.text:GetStringWidth() + 5), 0, 0)
		return b
	end
end

-- ApplyBackground
function module:ApplyBackground(frame)
	frame.bg = {}

	frame:SetMinResize(470, 260)

	frame.title = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.title:SetPoint("TOPLEFT", 10, -15)
	frame.title:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -10, -35)

	frame.bg.corner = frame:CreateTexture(nil, "OVERLAY")
	frame.bg.corner:SetTexture("Interface\\Addons\\Utopia\\Textures\\UtopiaCornerIcon")
	frame.bg.corner:SetWidth(59)
	frame.bg.corner:SetHeight(59)
	frame.bg.corner:SetPoint("TOPLEFT", 9, -6)

	frame:SetScript("OnSizeChanged",
		function(self, x, y)
			self:ArrangeTextures()
		end
	)

	function frame:ArrangeTextures()
		local h, w = self:GetHeight(), self:GetWidth()
		local rows = max(2, ceil(h / 256))
		local cols = max(2, ceil(w / 256))

		if (self.backgroundTextures) then
			for i,row in ipairs(self.backgroundTextures) do
				for j,tex in ipairs(row) do
					tex:Hide()
				end
			end
		else
			self.backgroundTextures = {}
		end

		local shrinkW = w < 512 and w / 2
		local shrinkH = h < 512 and h / 2
		local prevRow, prev

		for row = 1,rows do
			local rowTextures = self.backgroundTextures[row]
			if (not rowTextures) then
				rowTextures = {}
				self.backgroundTextures[row] = rowTextures
			end

			for col = 1,cols do
				local tex = rowTextures[col]
				if (not tex) then
					tex = frame:CreateTexture(nil, "BACKGROUND")
					rowTextures[col] = tex
				end
				tex:ClearAllPoints()
				tex:Show()

				if (col == 1 and row == 1) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopLeft")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, (shrinkH or 256) / 256)
					tex:SetPoint("TOPLEFT")

				elseif (row == 1 and col < cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-Top")
					if (col == cols - 1) then
						local sw = w % 256
						tex:SetWidth(sw)
						tex:SetTexCoord(0, sw / 256, 0, (shrinkH or 256) / 256)
					else
						tex:SetWidth(256)
						tex:SetTexCoord(0, 1, 0, (shrinkH or 256) / 256)
					end
					tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")
					tex:SetHeight(shrinkH or 256)
					
				elseif (row == 1 and col == cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopRight")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, (shrinkH or 256) / 256)
					tex:SetPoint("TOPRIGHT")

				elseif (row == rows and col == 1) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomLeft")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(0, (shrinkW or 256) / 256, 1 - (shrinkH or 256) / 256, 1)
					tex:SetPoint("BOTTOMLEFT")

				elseif (row == rows and col < cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-Bottom")
					if (col == cols - 1) then
						local sw = w % 256
						tex:SetWidth(sw)
						tex:SetTexCoord(0, sw / 256, 1 - (shrinkH or 256) / 256, 1)
					else
						tex:SetWidth(256)
						tex:SetTexCoord(0, 1, 1 - (shrinkH or 256) / 256, 1)
					end
					tex:SetPoint("BOTTOMLEFT", prev, "BOTTOMRIGHT")
					tex:SetHeight(shrinkH or 256)

				elseif (row == rows and col == cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetPoint("BOTTOMRIGHT")
					tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 1 - (shrinkH or 256) / 256, 1)

				elseif (row > 1 and row < rows) then
					if (col == 1) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomLeft")
						tex:SetTexCoord(0, 1, 0, 0.8)
						if (row == rows - 1) then
							local sh = h % 256
							tex:SetHeight(sh)
							tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, 0.8 / 256 * sh)
						else
							tex:SetHeight(256)
							tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, 0.8)
						end
						tex:SetWidth(shrinkW or 256)
						tex:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT")

					elseif (col == cols) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
						tex:SetTexCoord(0, 1, 0, 0.8)
						if (row == rows - 1) then
							local sh = h % 256
							tex:SetHeight(sh)
							tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, 0.8 / 256 * sh)
						else
							tex:SetHeight(256)
							tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, 0.8)
						end
						tex:SetWidth(shrinkW or 256)
						tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")

					elseif (col < cols) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
						local sw = w % 256
						local sh = h % 256
						local height, width = 256, 256
						local tch = 0.8
						local tcw = 0.8

						if (row == rows - 1) then
							height = sh
							tch = 0.8 / 256 * sh
						end
						if (col == cols - 1) then
							width = sw
							tcw = 0.8 / 256 * sw
						end

						tex:SetWidth(width)
						tex:SetHeight(height)
						tex:SetTexCoord(0, tcw, 0, tch)
						tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")
					end

				end

				if (col == 1) then
					prevRow = tex
				end
				prev = tex
			end
		end
	end

	frame.closeButton = self:CloseButton(frame)
end

-- GenericSuffixName
function module:GenericSuffixName(frame, suffix)
	local name, prefix
	if (frame:GetName()) then
		prefix = frame:GetName()
	else
		prefix = "Utopia"
	end
	if (not self.suffix) then
		self.suffix = {}
	end
	self.suffix[suffix] = (self.suffix[suffix] or 0) + 1
	return format("%s%s%d", prefix, suffix, self.suffix[suffix])
end

do
	local function onHide(self)
		self:GetParent():Hide()
	end
	function module:CloseButton(frame)
		local name = self:GenericSuffixName(frame, "CloseButton")
		local b = CreateFrame("Button", name, frame, "UIPanelCloseButton")
		b:SetScript("OnClick", onHide)
		b:SetPoint("TOPRIGHT", -3, -8)
		return b
	end
end

do
	-- listentry.OnClick
	local function entryOnClick(self)
		local p = self:GetParent()
		if (p.selectable) then
			local current = self:GetID() + p.offset
			if (p.multiselect and IsControlKeyDown()) then
				p.selected[current] = not p.selected[current] or nil

			elseif (p.multiselect and IsShiftKeyDown()) then
				for i = p.focus,current,(p.focus < current and 1 or -1) do
					local line = rawget(p.lines, i - p.offset)
					if (line) then
						line:LockHighlight()
					end
					p.selected[i] = true
				end
			else
				p.focus = current
				for i,line in pairs(p.lines) do
					line:UnlockHighlight()
				end
				wipe(p.selected)
				p.selected[current] = true
			end

			if (p.selected[current]) then
				self:LockHighlight()
			else
				self:UnlockHighlight()
			end
		end
		p:DoOnClick(self)
	end

	-- listentry.OnDoubleClick
	local function entryOnDoubleClick(self)
		local p = self:GetParent()
		if (p.OnDoubleClick) then
			p:OnDoubleClick(unpack(p.items[self.index], 2))
		end
	end

	-- list.DoOnSelectMultiple
	local function listDoOnSelectMultiple(self)
		if (self.OnSelectMultiple) then
			if (self.multiselect) then
				local temp = new()
				for id in pairs(self.selected) do
					tinsert(temp, id)
				end
				sort(temp)
				self:OnSelectMultiple(unpack(temp))
				del(temp)
			end
		end
	end

	-- list.DoOnClick
	local function listDoOnClick(self, from)
		if (self.multiselect and self.OnSelectMultiple) then
			self:DoOnSelectMultiple()
		elseif (self.OnClick) then
			self:OnClick(unpack(self.items[from.index], 2))
		end
	end

	-- list.Click
	local function listClick(self, ...)
		self:Select(...)
		local line = rawget(self.lines, (...) - self.offset)
		if (line) then
			self:DoOnClick(line)
		end
	end

	-- list.Select
	local function listSelect(self, ...)
		if (self.selectable) then
			if (not self.selected) then
				self.selected = new()
			else
				wipe(self.selected)
			end
			if (...) then
				local count = select("#", ...) 
				if (count > 0) then
					for i = 1,count do
						local index = select(i, ...)
						self.selected[index] = true
						local line = rawget(self.lines, index - self.offset)
						if (line) then
							line:LockHighlight()
						end
					end
				end
			else
				for i,line in pairs(self.lines) do
					line:UnlockHighlight()
				end
			end
		end
	end

	-- list.SetItems
	-- list must be a double depth table of display values and keys
	local function listSetItems(self, list)
--@debug@
		assert(type(list) == "table")
		assert(type(list[1]) == "table")
--@no-debug@

		deepDel(self.items)

		self.items = list
		self:Select()
		self:PopulateList()
	end

	-- list.Clear
	local function listClear(self)
		self.items = del(self.items)
		self.selected = del(self.selected)
	end

	-- list.PopulateList
	local function listPopulateList(self)
		if (not self.items) then
			return
		end

		local displayLines = floor(self:GetHeight() / 16)

		for i = 1,displayLines do
			local entryNo = i + self.offset
			if (entryNo > #self.items) then
				local b = rawget(self.lines, i)
				if (b) then
					b:Hide()
				end
			else
				local b = self.lines[i]
				b:SetText(self.items[entryNo][1])
				b:Show()
				b.index = entryNo

				if (self.selected and self.selected[entryNo]) then
					b:LockHighlight()
				else
					b:UnlockHighlight()
				end
			end
		end
		for i = displayLines + 1,#self.lines do
			self.lines[i]:Hide()
		end

		self.scroll:Update(#self.items)
	end

	-- list.OnVerticalScroll
	local function listOnVerticalScroll(self, offset)
		self.offset = offset
		self:PopulateList()
	end

	-- list.GetSelected
	local function listGetSelected(self)
		if (self.selectable and self.selected) then
			if (self.multiselect) then
				local temp = new()
				for id in pairs(self.selected) do
					tinsert(temp, id)
				end
				sort(temp)
				return unpack(temp)
			else
				return (next(self.selected))
			end
		end
	end

	-- list.OnUpdateDragging
	local function listOnUpdateDragging(self, elapsed)
		if (MouseIsOver(self)) then
			local dragSelectEnd
			for i = 1,#self.lines do
				local line = rawget(self.lines, i)
				if (line:IsShown() and MouseIsOver(line)) then
					dragSelectEnd = line:GetID() + self.offset
					break
				end
			end

			if (dragSelectEnd) then
				for i,line in pairs(self.lines) do
					line:UnlockHighlight()
				end
				wipe(self.selected)

				for i = self.dragSelectStart,dragSelectEnd,(self.dragSelectStart < dragSelectEnd and 1 or -1) do
					local line = rawget(self.lines, i - self.offset)
					if (line) then
						line:LockHighlight()
					end
					self.selected[i] = true
				end
			end

			self.moveDelay = 0.2
		else
			local x, y = GetCursorPosition()
			y = y / self:GetEffectiveScale()

			if (y > self:GetTop() or y < self:GetBottom()) then
				self.moveTimer = (self.moveTimer or 0) + elapsed
				if (self.moveTimer > self.moveDelay) then
					self.moveDelay = max(0.02, self.moveDelay * 0.95)
					self.moveTimer = 0

					if (y > self:GetTop()) then
						if (self.offset > 0) then
							self.scroll:SetOffset(self.offset - 1)
						end
					else
						local displayLines = floor(self:GetHeight() / 16)
						if (self.offset + displayLines < #self.items) then
							self.scroll:SetOffset(self.offset + 1)
						end
					end
				end
			else
				self.moveDelay = 0.2
			end
		end
	end

	-- listentry.OnDragStart
	local function entryOnDragStart(self)
		local p = self:GetParent()
		p.dragSelectStart = self:GetID() + p.offset
		p:SetScript("OnUpdate", listOnUpdateDragging)
	end

	-- listentry.OnDragStop
	local function entryOnDragStop(self)
		local p = self:GetParent()
		p:SetScript("OnUpdate", nil)
		self.dragSelectStart = nil
		p:DoOnSelectMultiple(self)
	end

	-- list.CreateListEntry
	local function CreateListEntry(parent, index, lines)
		local b = CreateFrame("Button", nil, parent)
		b:SetNormalFontObject(GameFontNormal)
		b:SetHighlightFontObject(GameFontHighlight)
		b:SetDisabledFontObject(GameFontDisable)
		b:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		b:SetText(" ") -- Required for GetFontString to work
		b.text = b:GetFontString()
		b.text:SetAllPoints()
		b.text:SetJustifyH("LEFT")

		b:SetID(index)
		if (index == 1) then
			b:SetPoint("TOPLEFT")
			b:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", 0, -16)
		else
			b:SetPoint("TOPLEFT", rawget(lines, index - 1), "BOTTOMLEFT")
			b:SetPoint("BOTTOMRIGHT", rawget(lines, index - 1), "BOTTOMRIGHT", 0, -16)
		end
		b:SetScript("OnClick", entryOnClick)
		b:SetScript("OnDoubleClick", entryOnDoubleClick)

		if (parent.multiselect) then
			b:RegisterForDrag("LeftButton")
			b:SetScript("OnDragStart", entryOnDragStart)
			b:SetScript("OnDragStop", entryOnDragStop)
		end

		return b
	end

	-- CreateListFrame
	function module:CreateListFrame(parent, selectable, multiselect)
		local name = self:GenericSuffixName(parent, "List")
		local list = CreateFrame("Frame", name, parent)
		list.selectable = selectable
		list.multiselect = multiselect
		list.offset, list.focus = 0, 1
		if (selectable or multiselect) then
			list.selected = {}
		end

		list.lines = setmetatable({}, {
			__index = function(self, index)
				assert(type(index) == "number")
				local b = CreateListEntry(list, index, self)
				self[index] = b
				return b
			end
		})

		list:SetScript("OnSizeChanged", listPopulateList)

		list.scroll				= self:CreateScrollFrame(list)
		list.Select				= listSelect
		list.GetSelected		= listGetSelected
		list.DoOnClick			= listDoOnClick
		list.DoOnSelectMultiple	= listDoOnSelectMultiple
		list.Click				= listClick
		list.PopulateList		= listPopulateList
		list.SetItems			= listSetItems
		list.OnVerticalScroll	= listOnVerticalScroll
		list.Clear				= listClear

		return list
	end
end

do
	-- scroll.Update
	local function scrollUpdate(self, total)
		local parent = self:GetParent()
		if (not self.parentWidth) then
			self.parentWidth = parent:GetWidth()
		end

		local visibleLines = floor(self:GetParent():GetHeight() / 16)
		local bars = parent.verticalBar
		local notWhere = parent.where == "LEFT" and "RIGHT" or "LEFT"
		if (FauxScrollFrame_Update(self, total, visibleLines, 16)) then
			parent:SetWidth(self.parentWidth - 21)
			self:Show()
			if (bars) then
				bars:ClearAllPoints()
				bars:SetPoint("TOP"..notWhere, self, "TOP"..bars.where, bars.where == "RIGHT" and 20 or -20, 2)
				bars:SetPoint("BOTTOM"..bars.where, self, "BOTTOM"..bars.where, bars.where == "RIGHT" and 28 or -28, 0)
			end
		else
			parent:SetWidth(self.parentWidth)
			self:Hide()
			if (bars) then
				bars:ClearAllPoints()
				bars:SetPoint("TOP"..notWhere, parent, "TOP"..bars.where, 0, 2)
				bars:SetPoint("BOTTOM"..bars.where, parent, "BOTTOM"..bars.where, bars.where == "RIGHT" and 8 or -8, 0)
			end
		end
	end

	-- onverticalscroll
	local function onverticalscroll(self, offset2)
		local list = self:GetParent()
		if (list.OnVerticalScroll) then
			list.offset = self.offset
			list:OnVerticalScroll(self.offset)
		end
	end

	-- scroll.OnVerticalScroll
	local function scrollOnVerticalScroll(self, offset)
		self.offset = offset
    	FauxScrollFrame_OnVerticalScroll(self, offset, 16, onverticalscroll)
	end

	-- scroll.SetOffset
	local function scrollSetOffset(self, offset)
		self.scrollbar:SetValue(max(0, offset) * self.scrollbar:GetValueStep())
	end

	-- CreateScrollFrame
	function module:CreateScrollFrame(frame)
		local name = self:GenericSuffixName(frame, "ScrollBar")
		local scroll = CreateFrame("ScrollFrame", name, frame, "FauxScrollFrameTemplate")
		scroll.scrollbar = _G[name.."ScrollBar"]
		scroll:Hide()
		scroll:SetPoint("TOPRIGHT", 1, 0)
		scroll:SetPoint("BOTTOMLEFT", 0, 2)
		scroll:SetScript("OnVerticalScroll", scrollOnVerticalScroll)
		scroll.Update = scrollUpdate

		scroll.tex1 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex1:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex1:SetWidth(31)
		scroll.tex1:SetHeight(64)
		scroll.tex1:SetPoint("TOPLEFT", scroll, "TOPRIGHT", -2, 3)
		scroll.tex1:SetTexCoord(0, 0.484375, 0, 0.25)

		scroll.tex2 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex2:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex2:SetWidth(31)
		scroll.tex2:SetHeight(106)
		scroll.tex2:SetPoint("BOTTOMLEFT", scroll, "BOTTOMRIGHT", -2, -2)
		scroll.tex2:SetTexCoord(0.515625, 1, 0, 0.4140625)

		scroll.tex3 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex3:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex3:SetPoint("TOPLEFT", scroll.tex1, "BOTTOMLEFT")
		scroll.tex3:SetPoint("BOTTOMRIGHT", scroll.tex2, "TOPRIGHT")
		scroll.tex3:SetTexCoord(0.515625, 1, 0, 0.3)

		scroll.SetOffset = scrollSetOffset

		return scroll
	end
end

do
	-- OnUpdate
	local function barOnUpdate(self, elapsed)
		local xCursor, yCursor = GetCursorPosition()
		xCursor, yCursor = xCursor / self:GetEffectiveScale(), yCursor / self:GetEffectiveScale()
		local xFrame, yFrame = self:GetCenter()

		local parentFrame = self:GetParent()
		if (self.vertical) then
			xCursor = max(xCursor, parentFrame:GetLeft() + 75)
			xCursor = min(xCursor, parentFrame:GetLeft() + 300)

			if (xCursor < xFrame - 1 or xCursor > xFrame + 1) then
				local point, parent, relpoint, x, y = parentFrame:GetPoint(2)
				x = x - (xFrame - xCursor)
				parentFrame:SetPoint(point, parent, relpoint, x, y)
			end
		else
			if (yCursor < yFrame - 1 or xCursor > yFrame + 1) then
				local point, parent, relpoint, x, y = parentFrame:GetPoint(2)
				y = y - (yFrame - yCursor)
				parentFrame:SetPoint(point, parent, relpoint, x, y)
			end
		end
	end

	-- OnDragStart
	local function barOnDragStart(self)
		self:SetScript("OnUpdate", barOnUpdate)
	end

	-- OnDragStop
	local function barOnDragStop(self)
		self:SetScript("OnUpdate", nil)
	end

	-- CreateVerticalBar
	function module:CreateVerticalBar(frame, where, draggable)
		assert(where == "LEFT" or where == "RIGHT")
		local bars = CreateFrame("Frame", nil, frame)
		bars.bars = {}
		bars.vertical = true
		bars.where = where
		frame.verticalBar = bars
		local notWhere = where == "LEFT" and "RIGHT" or "LEFT"

		bars:SetPoint("TOP"..notWhere, frame, "TOP"..where, 0, 2)
		bars:SetPoint("BOTTOM"..where, frame, "BOTTOM"..where, where == "RIGHT" and 8 or -8, 0)

		for i = 1,3 do
			local bar = bars:CreateTexture(nil, "OVERLAY")
			bar:SetTexture("Interface\\FriendsFrame\\UI-ChannelFrame-VerticalBar")
			tinsert(bars.bars, bar)

			if (i == 1) then
				bar:SetPoint("TOPLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0, 0.171875, 0, 0.171875)
			elseif (i == 2) then
				bar:SetPoint("BOTTOMLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0.82, 0.98, 0.798, 0.9609375)
			else
				bar:SetPoint("TOPLEFT", bars.bars[1], "BOTTOMLEFT")
				bar:SetPoint("BOTTOMRIGHT", bars.bars[2], "TOPRIGHT")
				bar:SetTexCoord(0.39, 0.6, 0, 1)
			end
		end

		if (draggable) then
			bars.highlight = bars:CreateTexture(nil, "HIGHLIGHT")
			bars.highlight:SetPoint("TOPLEFT", 3, 0)
			bars.highlight:SetPoint("BOTTOMRIGHT", -3, 0)
			bars.highlight:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
			bars.highlight:SetVertexColor(1, 1, 1, 0.5)

			bars:EnableMouse(true)
			bars:RegisterForDrag("LeftButton")
			bars:SetScript("OnDragStart", barOnDragStart)
			bars:SetScript("OnDragStop", barOnDragStop)
		end

		return bars
	end

	-- CreateHorizontalBar
	function module:CreateHorizontalBar(frame, where, draggable)
		assert(where == "TOP" or where == "BOTTOM")
		local bars = CreateFrame("Frame", nil, frame)
		bars.bars = {}
		bars.horizontal = true
		bars.where = where
		frame.horizontalBar = bars
		local notWhere = where == "TOP" and "BOTTOM" or "TOP"

		bars:SetPoint(notWhere.."LEFT", frame, where.."LEFT", -4, 0)
		bars:SetPoint(where.."RIGHT", frame, where.."RIGHT", 0, where == "TOP" and 8 or -8)

		for i = 1,3 do
			local bar = bars:CreateTexture(nil, "OVERLAY")
			bar:SetTexture("Interface\\ClassTrainerFrame\\UI-ClassTrainer-HorizontalBar")
			tinsert(bars.bars, bar)

			if (i == 1) then
				bar:SetPoint("TOPLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0, 0.25, 0, 0.25)
			elseif (i == 2) then
				bar:SetPoint("TOPRIGHT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0.09, 0.3, 0.25, 0.5)
			else
				bar:SetPoint("TOPLEFT", bars.bars[1], "TOPRIGHT")
				bar:SetPoint("BOTTOMRIGHT", bars.bars[2], "BOTTOMLEFT")
				bar:SetTexCoord(0.06, 1, 0, 0.25)
			end
		end

		if (draggable) then
			bars.highlight = bars:CreateTexture(nil, "HIGHLIGHT")
			bars.highlight:SetPoint("TOPLEFT", 0, -3)
			bars.highlight:SetPoint("BOTTOMRIGHT", 0, 3)
			bars.highlight:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
			bars.highlight:SetVertexColor(1, 1, 1, 0.5)

			bars:EnableMouse(true)
			bars:RegisterForDrag("LeftButton")
			bars:SetScript("OnDragStart", barOnDragStart)
			bars:SetScript("OnDragStop", barOnDragStop)
		end

		return bars
	end
end
