local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia", true)
if (not Utopia or Utopia.modules.UpTime) then
	return
end
Utopia:UpdateVersion("$Revision: 97 $")

local d = Utopia.debugprint
local module = Utopia:NewModule("UpTime", "AceHook-3.0", "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local gui = LibStub("AceGUI-3.0")
local bossid = LibStub("LibBossIDs-1.0").BossIDs
local db, dlookup, blookup, mlookup
local active
local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate, propercase = Utopia.rotate, Utopia.propercase
local band = bit.band

local BossIDWeakLookup = setmetatable({}, {
	__mode = "k",
	__index = function(self, guid)
		local id = tonumber(strsub(guid, -12, -7), 16)
		self[guid] = bossid[id] and true or false
		return self[guid]
	end,
})

module.options = {
	type = "group",
	name = L["Up-Time"],
	desc = L["Records debuff up-time on fights"],
	get = function(info) return module.db.profile[info[#info]] end,
	set = function(info, value) module.db.profile[info[#info]] = value end,
	guiInline = true,
	order = 100,
	args = {
		bossesOnly = {
			type = "toggle",
			name = L["Bosses Only"],
			desc = L["Only record boss fights"],
			order = 100,
		},
		notifications = {
			type = "toggle",
			name = L["Notifications"],
			desc = L["Display start and end statistics of fights. More to remind you that data is being stored and munching your addon ram up than anything else"],
			order = 110,
		},
		minimumDuration = {
			type = "range",
			name = L["Minimum Duration"],
			desc = L["Set the minimum duration for a fight to be stored for (last fight is always stored temporarily)"],
			order = 120,
			min = 0,
			max = 600,
			step = 1,
		},
		keep = {
			type = "range",
			name = L["History Size"],
			desc = L["How many fights to keep"],
			order = 130,
			min = 1,
			max = 500,
			step = 1,
		},
		recordDPS = {
			type = "toggle",
			name = L["Record DPS"],
			desc = L["Records DPS during fights. Just the basics; damage done to each mob per second by any source in your group. So you can see how raid debuffs are affecting things"],
			order = 140,
		},
		recordDeaths = {
			type = "toggle",
			name = L["Record Deaths"],
			desc = L["Records raid member deaths during fights"],
			order = 150,
		},
		display = {
			type = "group",
			name = L["Display"],
			desc = L["Display settings"],
			guiInline = true,
			get = function(info) return db[info[#info]] end,
			set = function(info, value) db[info[#info]] = value end,
			order = 300,
			args = {
				coloursExtreme = {
					type = "toggle",
					name = L["Coloured Categories"],
					desc = L["Colour the category names to reflect all the class colours applicable (Bring Shades)"],
					order = 10,
		        },
			},
		},
	},
}

local FRIEND_FLAGS = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
local ENEMY_FLAGS = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL
local EXLCUDE_FLAGS = COMBATLOG_OBJECT_TYPE_OBJECT + COMBATLOG_OBJECT_TYPE_PLAYER

-- MakeDestKey
-- Make a compressed Mob GUID for storing, and an ID -> Name lookup attached to the fight data
-- 0xF530000ABCD000123 becomes 43981:1
-- 0xF530000ABCD000492 becomes 43981:2
function module:MakeDestKey(fight, name, guid)
	local mobID = tonumber(guid:sub(-12, -7), 16)
	local mobUnique = tonumber(guid:sub(-6), 16)

	local lookup = self.fight.mobLookup[mobID]
	if (not lookup) then
		lookup = new()
		self.fight.mobLookup[mobID] = lookup
	end
	local key = lookup[mobUnique]
	if (not key) then
		key = (self.lastID[mobID] or 0) + 1
		self.lastID[mobID] = key
		lookup[mobUnique] = key
	end

	local mob = fight.mobs[mobID]
	if (not mob) then
		if (not name) then
			return
		end
		fight.mobs[mobID] = name
	end
	if (key == 1) then
		return mobID
	end
	return format("%d:%d", mobID, key)
end

-- MakeSourceKey
-- Make a compressed Player GUID for storing, and an ID -> Name lookup attached to the fight data
-- 0x0100000000009AFEB becomes 9AFEB
function module:MakeSourceKey(fight, name, guid)
	local playerID = tonumber(guid:sub(-8), 16)
	local player = fight.players[playerID]
	if (not player) then
		fight.players[playerID] = name
	end
	return format("%X", playerID)
end

-- BuildStoredData
function module:BuildStoredData(fight, timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId)
	local playerID = tonumber(sourceGUID:sub(-8), 16)
	local sourceKey = self:MakeSourceKey(fight, sourceName, sourceGUID)
	timeEnd = timeEnd and (timeEnd - timeStart)
	timeStart = max(0, timeStart - fight.combatStart)
	return format("%.3f,%.3f,%s,%d,%s", timeStart, timeEnd or 0, sourceKey, spellId, realSpellId or "")
end

-- GetStoredData
function module:GetStoredData(fight, data)
	local timeStart, timeEnd, sourceID, spellId, realSpellId = strsplit(",", data)
	sourceID = tonumber(sourceID, 16)
	timeStart = tonumber(timeStart) + fight.combatStart
	return timeStart, tonumber(timeEnd) + timeStart, fight.players[sourceID], format("0x010%013X", sourceID), tonumber(spellId), tonumber(realSpellId)
end

-- MobIDFromKey
local function MobIDFromKey(key)
	if (type(key) == "number") then
		return key
	else
		local mobid, index = strsplit(":", key)
		mobid = tonumber(mobid)
		index = index and tonumber(index) or nil
		return mobid, index
	end
end

-- MakeMobGUID
local function MakeMobGUID(key)
	local id, ind = MobIDFromKey(key)
	return format("0xF53%07X%06X", id, ind or 1)
end

-- AddAura
function module:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, realSpellId, realSpellName, auraType)
	if (destName and sourceName and dlookup[spellName]) then
		if (not self.fight.combatStart) then
			self.fight.combatStart = timestamp
		end

		local target = active[destGUID]
		if (not target) then
			target = new()
			active[destGUID] = target
		end

		local source = target[sourceGUID]
		if (not source) then
			source = new()
			target[sourceGUID] = source
		end

		local s = source[spellName]
		if (not s) then
			local n = new()
			n.timeStart = timestamp
			n.spellId = spellId
			n.realSpellId = realSpellId

			source[spellName] = n
		end

		self.lastEventTime = timestamp
		self.applyCount = self.applyCount + 1
	end

	local mapList = mlookup[spellName]
	if (mapList) then
		for i,map in ipairs(mapList) do
			local mapping = dlookup[map.key]
			if (not mapping or not mapping.requiredTalent or Utopia:HasTalent(sourceName, mapping.requiredTalent)) then
				self:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, mapping and mapping.id or 0, map.key, realSpellId or spellId, realSpellName or spellName, auraType)
			end
		end
	end
end

-- RemoveAura
function module:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, realSpellId, realSpellName, auraType)
	if (dlookup[spellName]) then
		local target = active[destGUID]
		if (target) then
			local source = target[sourceGUID]
			if (source) then
				local s = source[spellName]
				if (s) then
					source[spellName] = nil

					local fight = self.fight
					if (not fight.combatStart) then
						fight.combatStart = timestamp
					end

					if (timestamp > s.timeStart + 0.0009) then	-- Skip things that disappear the same timestamp they apply
						local entry = self:BuildStoredData(fight, s.timeStart, timestamp, sourceName, sourceGUID, spellId, realSpellId)
						local destKey = self:MakeDestKey(fight, destName, destGUID)
						if (destKey) then
							local data = fight.data[destKey]
							if (not data) then
								data = new()
								fight.data[destKey] = data
							end
							tinsert(data, entry)
						end
					end

					self.lastEventTime = timestamp
					self.removeCount = self.removeCount + 1
					del(s)
				end
			end
		end
	end

	local mapList = mlookup[spellName]
	if (mapList) then
		for i,map in ipairs(mapList) do
			local mapping = dlookup[map.key]
			if (not mapping or not mapping.requiredTalent or Utopia:HasTalent(sourceName, mapping.requiredTalent)) then
				module:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, mapping and mapping.id or 0, map.key, realSpellId or spellId, realSpellName or spellName, auraType)
			end
		end
	end
end

-- RecordDamage
function module:RecordDamage(timestamp, raidDmg, guid, name, amount)
	local f = self.fight
	if (not f.combatStart) then
		f.combatStart = timestamp
	end

	local key = self:MakeDestKey(f, name, guid)
	if (key) then
		local index = floor(timestamp - f.combatStart) + 1

		local target
		if (raidDmg) then
			target = self.dps[key]
			if (not target) then
				target = new()
				self.dps[key] = target
			end
		else
			target = self.bossdps[key]
			if (not target) then
				target = new()
				self.bossdps[key] = target
			end
		end			

		for i = #target + 1,index - 1 do
			tinsert(target, 0)
		end
		target[index] = (target[index] or 0) + amount
	end
end

local clEvents = {}
local dpsEvents = {}

-- SPELL_AURA_APPLIED
function clEvents:SPELL_AURA_APPLIED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
		if (band(destFlags, ENEMY_FLAGS) ~= 0 and band(destFlags, FRIEND_FLAGS) == 0) then
			if (band(destFlags, EXLCUDE_FLAGS) == 0) then
				if (not db.bossesOnly or BossIDWeakLookup[destGUID]) then
					self:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, nil, nil, auraType)
				end
			end
		end
	end
end

-- SPELL_AURA_REMOVED
function clEvents:SPELL_AURA_REMOVED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
		if (band(destFlags, ENEMY_FLAGS) ~= 0 and band(destFlags, FRIEND_FLAGS) == 0) then
			if (band(destFlags, EXLCUDE_FLAGS) == 0) then
				if (not db.bossesOnly or BossIDWeakLookup[destGUID]) then
					self:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, nil, nil, auraType)
				end
			end
		end
	end
end

-- UNIT_DIED
function clEvents:UNIT_DIED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags)
	if (self.deaths) then
		if (band(destFlags, FRIEND_FLAGS) ~= 0) then
			tinsert(self.deaths, {timestamp, sourceName, sourceGUID})
		end
	end
end

-- SPELL_DAMAGE
function dpsEvents:SWING_DAMAGE(timestamp, raidDmg, destGUID, destName, amount, overkill)
	self:RecordDamage(timestamp, raidDmg, destGUID, destName, amount - overkill)
end

-- SPELL_DAMAGE
function dpsEvents:SPELL_DAMAGE(timestamp, raidDmg, destGUID, destName, spellId, spellName, spellSchool, amount, overkill)
	self:RecordDamage(timestamp, raidDmg, destGUID, destName, amount - overkill)
end
dpsEvents.SPELL_PERIODIC_DAMAGE = dpsEvents.SPELL_DAMAGE

local ignoreSpells = {
	[GetSpellInfo(53338)] = true,				-- Hunter's Mark
	[GetSpellInfo(49005)] = true,				-- Mark of Blood
}

-- COMBAT_LOG_EVENT_UNFILTERED
function module:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	if (not active) then
		-- Check for start of a fight by someone in group, first events will happen before you enter combat
		-- We do this because first events will often happen before you're in combat, and we don't really
		-- want to miss anything
		if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
			if (band(destFlags, ENEMY_FLAGS) ~= 0 and band(destFlags, FRIEND_FLAGS) == 0) then
				if (band(destFlags, EXLCUDE_FLAGS) == 0) then
					if (UnitInRaid(sourceName) or UnitInParty(sourceName)) then
						if (event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REMOVED" or event == "SPELL_AURA_REFRESH") then
							local spellId, spellName, spellSchool, auraType = ...
							if (ignoreSpells[spellName]) then
								-- Don't start a fight with Hunter's Mark or Mark of Blood
								return
							end
						end

						self:StartFighting()
					end
				end
			end
		end
	end
	if (active) then
		local f = clEvents[event]
		if (f) then
			f(self, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
		elseif (self.dps) then
			f = dpsEvents[event]
			if (f) then
				if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
					if (band(destFlags, ENEMY_FLAGS) ~= 0 and band(destFlags, FRIEND_FLAGS) == 0) then
						if (not db.bossesOnly or BossIDWeakLookup[destGUID]) then
							f(self, timestamp, true, destGUID, destName, ...)
						end
					end
				elseif (band(sourceFlags, ENEMY_FLAGS) ~= 0 and band(sourceFlags, EXLCUDE_FLAGS) == 0) then
					if (band(destFlags, FRIEND_FLAGS) ~= 0) then
						if (not db.bossesOnly or BossIDWeakLookup[sourceGUID]) then
							f(self, timestamp, false, sourceGUID, sourceName, ...)
						end
					end					
				end
			end
		end
	end
end

-- PLAYER_REGEN_DISABLED
function module:PLAYER_REGEN_DISABLED()
	if (not Utopia:ShouldShow()) then
		return
	end
	self:StartFighting()
end

-- StartFighting
function module:StartFighting()
	self:EndFighting()

	--d("Fight Started: %s", date("%X",time()))
	self.active = {}
	self.fight = {}
	self.fight.boss = self.activeBossEncounter
	self.fight.data = {}
	self.fight.players = {}
	self.fight.mobs = {}
	--self.fight.combatStart = time()
	self.lastID = {}
	self.fight.mobLookup = {}
	if (db.recordDPS) then
		self.dps = {}
		self.bossdps = {}
	end
	if (db.recordDeaths) then
		self.deaths = {}
	end

	active = self.active

	self.combatStart = time()
	self.applyCount, self.removeCount = 0, 0

	dlookup = Utopia.lookup.debuffs
	blookup = Utopia.lookup.buffs
	mlookup = Utopia.mappingLookup
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self:CheckEndOfFight()
end

-- PLAYER_REGEN_ENABLED
function module:PLAYER_REGEN_ENABLED()
	if (not self:CheckEndOfFight()) then
		self:RegisterEvent("UNIT_DIED")
		if (not self.fightEndTimer) then
			self.fightEndTimer = self:ScheduleRepeatingTimer("CheckEndOfFight", 2)
		end
	end
end

-- UNIT_DIED
function module:UNIT_DIED()
	self:CheckEndOfFight()
end

-- CheckEndOfFight
function module:CheckEndOfFight()
	if (not InCombatLockdown()) then
		for unit in Utopia:IterateRoster() do
			if (UnitAffectingCombat(unit)) then
				return
			end
		end
		self:EndFighting()
		return true
	end
end

-- EndFighting
function module:EndFighting()
	if (self.fightEndTimer) then
		self:CancelTimer(self.fightEndTimer)
		self.fightEndTimer = nil
	end
	if (not self.fight) then
		return
	end

	--d("Fight Ended: %s", date("%X",time()))
	self:UnregisterEvent("UNIT_DIED")
	self:UnregisterEvent("PLAYER_LEAVING_WORLD")

	if (not db.history) then
		db.history = {}
	end

	self:Cleanup()

	if (self.fight.combatStart and time() - self.fight.combatStart >= db.minimumDuration) then
		if (next(self.fight.data)) then
			local fight = self.fight
			self.fight = nil

			fight.mobLookup = deepDel(fight.mobLookup)
			fight.combatEnd = self.lastEventTime
			fight.player = UnitName("player")

			tinsert(db.history, fight)

			if (db.notifications) then
				local link = format("|cFF80FF80|Hutopia_fight:%d|h[%s]|h|r", fight.combatStart, L["View"])
				Utopia:Print(format(L["Recorded |cFFFFFF80%s|r (%d secs) - |cFF80FF80%d|r auras %s"], date("%X"), time() - fight.combatStart, max(self.applyCount, self.removeCount), link))
			end

			if (self.dps and next(self.dps)) then
				fight.dps = {}
				for mobKey,data in pairs(self.dps) do
					fight.dps[mobKey] = table.concat(data, ",")
				end
			end
			if (self.bossdps and next(self.bossdps)) then
				fight.bossdps = {}
				for mobKey,data in pairs(self.bossdps) do
					fight.bossdps[mobKey] = table.concat(data, ",")
				end
			end
			if (self.deaths and #self.deaths > 0) then
				for i,info in ipairs(self.deaths) do
					local key = self:MakeSourceKey(f, info[2], info[3])
					self.deaths[i] = strjoin(",", timestamp - fight.combatStart, key)
				end
				fight.deaths = table.concat(info, ";")
			end

			self.fight = nil
		end
	end

	self.dps = deepDel(self.dps)
	self.bossdps = deepDel(self.bossdps)
	self.deaths = deepDel(self.deaths)
	self.active = deepDel(self.active)
	self.fight = deepDel(self.fight)
	self.lastID = deepDel(self.lastID)
	active = nil
end

-- SetItemRef
function module:SetItemRef(link, text, button)
	local cmd, when = strsplit(":", link)
	if (cmd == "utopia_fight") then
		when = tonumber(when)
		for i,fight in ipairs(db.history) do
			if (fight.combatStart == when) then
				module:OpenView(i)
				break
			end
		end
		return true
	end
	return self.hooks.SetItemRef(link, text, button)
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	if (UnitExists("target")) then
		local name = UnitName("target")
		local classification = UnitClassification("target")
		self.key = name.."-"..date()
	end
end

-- Cleanup
function module:Cleanup()
	local purged = 0
	for i = #db.history,1,-1 do
		local fight = db.history[i]
		if (not fight.keep and fight.combatEnd - fight.combatStart < db.minimumDuration) then
			purged = purged + 1
			deepDel(tremove(db.history, i))
		end
	end

	for i = #db.history,1,-1 do
		if (#db.history <= db.keep) then
			break
		end

		local fight = db.history[i]
		if (not fight.keep) then
			purged = purged + 1
			deepDel(tremove(db.history, i))
		end
	end
end

-- GetUptime
-- Get's the uptime of a spell, defaulting to the last fight
--[[
function module:GetUptime(spellName)
	local data = db.history
	if (data) then
		data = data[#data]
		local spells = data and data.spell and data.spell[spellName]
		if (spells) then
			local linear = data and data.linear
			if (linear) then
				local startTime = data.combatStart
				local endTime = data.combatEnd
				local on, off = 0, 0
				for i,info in ipairs(spells) do
					on = on + (info.timeStart - startTime)
					if (info.timeEnd) then
						off = off + (info.timeEnd - startTime)
					else
						off = off + (endTime - startTime)
					end
				end

				on = on / #spells
				off = off / #spells

				local buffTime = off - on
				local fightTime = endTime - startTime

				return buffTime, fightTime
			end
		end
	end
end ]]

-- OpenView
function module:OpenView(key)
	if (not self.freeFrames) then
		self.freeFrames = {}
		self.frames = {}
	end
	local frame = tremove(self.freeFrames) or self:CreateUpTimeFrame()
	self.frames[frame] = true

	frame:SetView(key)
	return frame
end

do
	-- OnHide
	local function frameOnHide(self)
		self.offset = nil
		self.fight = nil
		self.area.fight = nil
		self.area:OnClose()
		self:ClearSelected()
		if (self.dpsFrame) then
			self.showDPS, self.showDPSBoss = nil, nil
			self:ShowDPS()
		end

		module.frames[self] = nil
		tinsert(module.freeFrames, self)

		self.orderedMobNames = del(self.orderedMobNames)
	end

	-- SetTitle
	local function frameSetTitle(self)
		local when = module:DescribeDate(self.fight.combatStart)
		local dur = self.fight.combatEnd - self.fight.combatStart
		if (dur > 60) then
			dur = format(MINUTES_ABBR.." "..SECONDS_ABBR, dur / 60, dur % 60)
		else
			dur = format(SECONDS_ABBR, dur)
		end
		self.title:SetFormattedText("%s: %s %s - %s [%s]", Utopia:ColourPlayer(self.fight.player), when, date("%X", self.fight.combatStart), date("%X", self.fight.combatEnd), dur)
	end

	-- OnShow
	local function frameOnShow(self)
		self:SetTitle()
		self.area:ShowScale()
	end

	-- framelist.OnClick
	local function framelistOnClick(self, what)
		local parent = self:GetParent()
		parent:ClearSelected()
		parent.area.mobKeys[what] = true

		local lowTime, highTime = parent.area:SelectMobKey()
		parent.area.rightScroll.targetTime = lowTime
		parent.area.rightScroll.targetTimeHigh = highTime
	end

	-- framelist.OnSelectMultiple
	local function framelistOnSelectMultiple(self, ...)
		local parent = self:GetParent()
		parent:ClearSelected()

		for i = 1,select("#",...) do
			local index = select(i, ...)
			parent:AddSelected(index)
		end

		local lowTime, highTime = parent.area:SelectMobKey()
		parent.area.rightScroll.targetTime = lowTime
		parent.area.rightScroll.targetTimeHigh = highTime
	end

	-- frame.ClearSelected
	local function frameClearSelected(self)
		del(self.area.mobKeys)
		self.area.mobKeys = new()
	end

	-- frame.AddSelected
	local function frameAddSelected(self, index)
		local name = self.orderedMobNames[index]
		local key = self.mobnameToKeyLookup[name]
		self.area.mobKeys[key] = true
	end

	-- frame.SetView
	local function frameSetView(self, key)
		self.key = key
		local fight = db.history[key]
		assert(fight)
		self.fight = fight
		self.area.fight = fight

		self:Show()

		self:ClearSelected()
		self:OrderMobNames()
		self:PopulateMobNames()
		self.list:Click(1)
	end

	-- frame.PopulateMobNames
	local function framePopulateMobNames(self)
		local listItems = new()
		for i,name in ipairs(self.orderedMobNames) do
			tinsert(listItems, new(name, self.mobnameToKeyLookup[name]))
		end
		self.list:SetItems(listItems)
	end

	-- nameSort, aware of numbers in braces
	local function nameSort(a, b)
		local name1, num1 = strmatch(a, "(.+) %((%d+)%)$")
		if (num1) then
			local name2, num2 = strmatch(b, "(.+) %((%d+)%)$")
			if (num2 and name1 == name2) then
				return tonumber(num1) < tonumber(num2)
			end
		end
		return a < b
	end

	-- OrderMobNames
	local function frameOrderMobNames(self)
		self.offset = 0
		self.orderedMobNames = del(self.orderedMobNames)
		self.mobnameToKeyLookup = del(self.mobnameToKeyLookup)
		self.orderedMobNames = new()
		self.mobnameToKeyLookup = new()
		local l = self.mobnameToKeyLookup
		local fight = self.fight
		local mobs = self.fight.mobs

		for mobKey,info in pairs(fight.data) do
			local mobid, index = MobIDFromKey(mobKey)
			local baseName = mobs[mobid]
			local name

			if (index and index > 1) then
				name = format("%s (%d)", baseName, index)
			else
				name = baseName
			end
			if (l[name]) then
				-- Two mobs of same name can have a different ID, so we run into the problem of duplicates here
				-- ie: Tempest Minion with Emalon; the spawned ones during fight have a different ID than the 4
				-- static ones at start
				for i = 2,10000 do
					local temp = format("%s (%d)", baseName, i)
					if (not l[temp]) then
						name = temp
						break
					end
				end
			end

			tinsert(self.orderedMobNames, name)
			l[name] = mobKey
		end
		sort(self.orderedMobNames, nameSort)
	end

	local function setModeCat(self)
		self:GetParent().area:SetMode("category")
	end
	local function setModeSpell(self)
		self:GetParent().area:SetMode("spell")
	end

	local function getPlayerNames(self)
		return self:GetParent().area.playerNames
	end

	local function setPlayerNames(self, newval)
		self:GetParent().area.playerNames = newval
		self:GetParent().area:SelectMobKey()
	end

	local function getShowDPS(self)
		local frame = self:GetParent()
		self:SetEnabled(frame.fight.dps)
		return frame.fight.dps and frame.showDPS
	end

	local function setShowDPS(self, newval)
		self:GetParent().showDPS = newval
		self:GetParent():ShowDPS()
	end

	local function getShowDPSBoss(self)
		local frame = self:GetParent()
		self:SetEnabled(frame.fight.bossdps)
		return frame.fight.bossdps and frame.showDPSBoss
	end

	local function setShowDPSBoss(self, newval)
		self:GetParent().showDPSBoss = newval
		self:GetParent():ShowDPS()
	end

	local function getShowDeaths(self)
		local frame = self:GetParent()
		self:SetEnabled(frame.fight.deaths)
		return frame.fight.deaths and frame.showDeaths
	end

	local function setShowDeaths(self, newval)
		self:GetParent().showDeaths = newval
		self:GetParent().area:PopulateDeaths()
	end

	local function frameGetKeep(self)
		local frame = self:GetParent()
		return frame.fight and frame.fight.keep
	end

	local function frameSetKeep(self, newval)
		local frame = self:GetParent()
		frame.fight.keep = newval and true
	end

	-- frame.ShowDPS
	local function frameShowDPS(self)
		if (self.showDPS or self.showDPSBoss) then
			if (not self.dpsFrame) then
				self.dpsFrame = module:CreateDPSFrame(self)
				self.horizontalBar = module:CreateHorizontalBar(self.dpsFrame, "TOP")
				self.horizontalBar:SetPoint("TOPRIGHT", self.dpsFrame, "TOPRIGHT", 24, 8)

				self.dpsFrame:SetPoint("BOTTOMRIGHT", -40, 43)
			else
				self.dpsFrame:Show()
			end
			self.dpsFrame:SetPoint("TOPLEFT", self.list.verticalBar, "BOTTOMRIGHT", 0, self.dpsFrame.height)
			self.dpsFrame:SetFight(self.fight)

			self.area:SetPoint("BOTTOMRIGHT", self.dpsFrame, "TOPRIGHT", 0, 6)
		else
			if (self.dpsFrame) then
				self.dpsFrame:Hide()
				self.dpsFrame:SetPoint("TOPLEFT", self.list, "BOTTOMRIGHT", 0, 0)
				self.area:SetPoint("BOTTOMRIGHT", self.dpsFrame, "TOPRIGHT", 0, 0)
			end
		end
	end

	-- CreateUpTimeFrame
	function module:CreateUpTimeFrame()
		self.frameNo = (self.frameNo or 0) + 1
		local frame = CreateFrame("Frame", "UtopiaUpTimeFrame"..self.frameNo, UIParent)
		frame:Hide()
		frame:SetScale(0.8)
		frame:SetWidth(768)
		frame:SetHeight(512)
		self:ApplyBackground(frame)
		self:MovableFrame(frame)

		frame.list = self:CreateListFrame(frame, true, true)
		frame.list:SetPoint("TOPLEFT", 24, -82)
		frame.list:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", 195, 42)
		frame.list.OnSelectMultiple = framelistOnSelectMultiple
		frame.list.OnClick = framelistOnClick

		frame.bar = self:CreateVerticalBar(frame.list, "RIGHT", true)

		frame.button1 = self:Button(frame, L["By Category"], setModeCat, "BOTTOMLEFT", 15, 15)
		frame.button2 = self:Button(frame, L["By Spell"], setModeSpell, "LEFT", 10, 0, frame.button1, "RIGHT")
		frame.checkbox1 = self:CheckBox(frame, L["Player Names"], getPlayerNames, setPlayerNames, "LEFT", 10, 0, frame.button2, "RIGHT")
		frame.dpsCheckbox = self:CheckBox(frame, L["Show DPS"], getShowDPS, setShowDPS, "LEFT", frame.checkbox1.text:GetStringWidth() + 15, 0, frame.checkbox1, "RIGHT")
		frame.dpsBossCheckbox = self:CheckBox(frame, L["Show Enemy DPS"], getShowDPSBoss, setShowDPSBoss, "LEFT", frame.dpsCheckbox.text:GetStringWidth() + 15, 0, frame.dpsCheckbox, "RIGHT")
		frame.deathsCheckbox = self:CheckBox(frame, L["Show Deaths"], getShowDeaths, setShowDeaths, "LEFT", frame.dpsBossCheckbox.text:GetStringWidth() + 15, 0, frame.dpsBossCheckbox, "RIGHT")

		local area = self:CreateWorkArea(frame)
		frame.area = area
		area:SetPoint("TOPLEFT", frame.list.verticalBar, "TOPRIGHT", 0, 0)
		area:SetPoint("BOTTOMRIGHT", -40, 43)

		-- Keep option
		frame.keep = module:CheckBox(frame, L["Keep"], frameGetKeep, frameSetKeep, "TOPLEFT", 75, -35)
		frame.keep.tooltip = L["Do not automatically purge this fight when it expires past the present age/size limits defined in options"]

		frame.OnShow			= frameOnShow
		frame.OnHide			= frameOnHide
		frame.SetTitle			= frameSetTitle
		frame.OrderMobNames		= frameOrderMobNames
		frame.PopulateMobNames	= framePopulateMobNames
		frame.SetView			= frameSetView
		frame.ClearSelected		= frameClearSelected
		frame.AddSelected		= frameAddSelected
		frame.ShowDPS			= frameShowDPS
		frame.ShowDeaths		= frameShowDeaths

		frame:SetScript("OnShow", frame.OnShow)
		frame:SetScript("OnHide", frame.OnHide)

		return frame
	end
end

do
	local areaCreateSpellLine

	-- area.Init
	local function areaInit(self)
		self:ClearDisplay()
		self.nameToLineLookup = new()
		self.dataIndex = new()
		self:BuildSpellNameList()
	end

	-- area.OnClose
	local function areaOnClose(self)
		self.scale = 6
		self.playerNames = nil
		self.lastTooltipInfo = del(self.lastTooltipInfo)
		self:ClearDisplay()
		self.rightScroll:SetHorizontalScroll(0)
		self.rightScroll:SetVerticalScroll(0)
		self.leftScroll:SetVerticalScroll(0)
	end

	-- area.SetMode
	local function areaSetMode(self, mode)
		self.mode = mode
		self:Init()
		self:Populate()
	end

	-- area.SelectMobKey
	local function areaSelectMobKey(self)
		self:Init()
		local lowTime, highTime = self:Populate()
		return lowTime, highTime
	end

	-- area.RightScrollOnUpdate
	local function areaRightScrollOnUpdate(self, elapsed)
		self:GetParent():ShowCursorLocation()
	end

	local rightscrollOnUpdateSwoosh
	-- rightscrollOnUpdateCheckScrollRange
	local function rightscrollOnUpdateCheckScrollRange(self, elapsed)
		self.onUpdateCheck = self.onUpdateCheck + 1
		if (self.onUpdateCheck > 1) then
			local area = self:GetParent()
			if (self.targetTime) then
				local w = self:GetWidth()
				local wTime = w / area.scale

				local range = self.targetTimeHigh - self.targetTime
				local spare = (range < wTime) and ((wTime - range) / 2) or 0
				self.swooshTarget = max(0, (self.targetTime - spare - area.fight.combatStart) * area.scale)
				self:SetScript("OnUpdate", rightscrollOnUpdateSwoosh)
				self.targetTime, self.targetTimeHigh = nil, nil
				return
			end

			-- Can only validate the dimensions of the ScrollChild after 2nd frames are drawn
			local offset = -self:GetHorizontalScroll()
			local newoffset = max(0, min(self:GetHorizontalScrollRange(), offset))
			if (offset ~= newoffset) then
				self:SetHorizontalScroll(-newoffset)
			end
			if (area:MouseInArea()) then
				self:SetScript("OnUpdate", areaRightScrollOnUpdate)
			else
				self:SetScript("OnUpdate", nil)
			end

			local area = self:GetParent()
			area.lastX, area.lastY = nil, nil
			area:ShowCursorLocation()
		end
	end

	-- rightscrollOnUpdateSwoosh
	function rightscrollOnUpdateSwoosh(self, elapsed)
		local offset = -self:GetHorizontalScroll()
		local diff = self.swooshTarget - offset
        local move = diff / 10 * 60 * elapsed
        local set = offset + move
        local Max = self:GetHorizontalScrollRange()
        if (self.swooshTarget > Max) then
        	self.swooshTarget = Max
        end

		if (set >= Max) then
			set = Max
			self.swooshTarget = nil
		elseif (set <= 0) then
			set = 0
			self.swooshTarget = nil
		elseif (abs(set - offset) < 0.1) then
			set = self.swooshTarget
			self.swooshTarget = nil
		end

		self:SetHorizontalScroll(-set)

		if (not self.swooshTarget) then
			self.targetTime, self.targetTimeHigh = nil, nil
			self.onUpdateCheck = 0
			self:SetScript("OnUpdate", rightscrollOnUpdateCheckScrollRange)
		end

		self:GetParent():ShowCursorLocation()
	end

	-- area.ResetLines
	local function areaResetLines(self)
		for i,line in ipairs(self.lines) do
			line:Hide()
			line:Show()
		end
	end

	-- area.ClearDisplay
	local function areaClearDisplay(self)
		self.shownLines = 0
		self.offset = nil
		self.dataIndex = del(self.dataIndex)
		self.nameToLineLookup = deepDel(self.nameToLineLookup)
		for i,line in ipairs(self.lines) do
			line.info = nil
			line.spellId = nil
			line:UnlockHighlight()
			line:Hide()
		end
		self.offset = 0
		self.rowHighlight = nil
		self:CleanupDeaths()
	end

	-- area.RightScrollOnEnter
	local function areaRightScrollOnEnter(self)
		if (not self.dragging and not self.speedX and not self.swooshTarget and not self.targetTime) then
			self:SetScript("OnUpdate", areaRightScrollOnUpdate)
		end
	end

	-- area.RightScrollOnLeave
	local function areaRightScrollOnLeave(self)
		local area = self:GetParent()
		if (not self.dragging and not self.speedX and not self.swooshTarget and not self.targetTime) then
			self:SetScript("OnUpdate", nil)
			area:HideCursorLocation()
		end
		area:RemoveHighlight()
		GameTooltip:Hide()
	end

	-- area.Populate
	local function areaPopulate(self)
		if (not self.mobKeys) then
			return
		end

		local lowTime, highTime = self:PopulateAuras()
		self:PopulateDeaths()

		return lowTime, highTime
	end

	-- area.PopulateAuras
	local function areaPopulateAuras(self)
		local fight = self.fight
		local data = fight.data
		local w = max(200, (self.fight.combatEnd - self.fight.combatStart) * self.scale)
		self.rightScrollChild:SetWidth(w)

		local parent = self:GetParent()
		local keyLookup = parent.mobnameToKeyLookup
		local lowTime, highTime

		for i,mobName in ipairs(parent.orderedMobNames) do
			local mobKey = keyLookup[mobName]

			if (self.mobKeys[mobKey]) then
				local fightData = data[mobKey]
				local index = self.dataIndex[mobKey] or 1

				for i = index,#fightData do
					local segment = fightData[i]
					local timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId = module:GetStoredData(fight, segment)
					if (timeEnd > timeStart + 0.0009) then		-- Sanity
						self:AddSpell(timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey)
					end
					if (not lowTime or timeStart < lowTime) then
						lowTime = timeStart
					end
					if (not highTime or timeEnd > highTime) then
						highTime = timeEnd
					end
				end

				self.dataIndex[mobKey] = #fightData + 1
			end
		end

		self.rightScroll.onUpdateCheck = 0
		self.rightScroll:SetScript("OnUpdate", rightscrollOnUpdateCheckScrollRange)

		if ((parent.showDPS or parent.showDPSBoss) and parent.dpsFrame) then
			parent.dpsFrame:SetFight(self.fight)
		end

		self:ShowScroll()

		return lowTime, highTime
	end

	-- area.PopulateDeaths
	local function areaPopulateDeaths(self)
		self:CleanupDeaths()
		local fight = self.fight
		if (not fight.deaths) then
			return
		end

		local parent = self:GetParent()
		if (not parent.showDeaths) then
			return
		end

		local w = max(200, (self.fight.combatEnd - self.fight.combatStart) * self.scale)

		local list = new(strsplit(";", fight.deaths))
		for i,str in ipairs(list) do
			local timeOffset, playerID = strsplit(",", str)
			if (playerID) then
				timeOffset, playerID = tonumber(timeOffset), tonumber(playerID)
				local player = fight.players[playerID]
				if (player) then
					local timestamp = fight.combatStart + timeOffset

					local death = self:AquireDeathMarker()
					death:SetVertexColor(1, 0, 0)
					death:SetPoint("TOPLEFT", timeOffset * self.scale, 0)
					death.text:SetText(player)
					death.text:SetTextColor(1, 0, 0)

					self.deaths = self.deaths or new()
					tinsert(self.deaths, death)
				end
			end
		end
	end

	local function areaCleanupDeaths(self)
		if (self.deaths) then
			for i,death in ipairs(self.deaths) do
				self:ReleaseDeathMarker(death)
			end
		end
		self.deaths = del(self.deaths)
	end

	local markers = {}
	local function areaCreateDeathMarker(self)
		local tex = self.rightScrollChild:CreateTexture(nil, "OVERLAY")
		tex:SetWidth(0.8)
		tex:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		tex.text = self.rightScrollChild:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		tex.text:SetPoint("TOP", tex, "BOTTOM", 0, 0)
		return tex
	end
	
	local function areaAquireDeathMarker(self, parent)
		local marker = tremove(markers) or self:CreateDeathMarker()
		marker:SetParent(self.rightScrollChild)
		marker.text:SetParent(self.rightScrollChild)
		marker:Show()
		marker.text:Show()
		marker:SetHeight(20 * self.shownLines + 5)
		return marker
	end

	local function areaReleaseDeathMarker(self, marker)
		tinsert(markers, marker)
		marker:Hide()
		marker.text:Hide()
		marker:ClearAllPoints()
		marker:SetParent(UIParent)
		marker.text:SetParent(UIParent)
	end

	-- area.CreateSlidingIndicator
	local function areaCreateSlidingIndicator(self)
		local pointer = self:CreateTexture(nil, "OVERLAY")
		pointer:SetTexture("Interface\\Addons\\Utopia\\Textures\\SlidingIndicator")
		pointer:SetWidth(24)
		pointer:SetHeight(8)
		pointer:SetTexCoord(0, 1, 0.05, 0.5)

		local text = self:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		text:SetPoint("BOTTOM", pointer, "TOP", 0, -3)

		pointer.text = text
		return pointer
	end

	-- area.HideCursorLocation
	local function areaHideCursorLocation(self)
		if (self.cursorTick) then
			self.cursorTick:Hide()
			self.cursorTick.text:Hide()
		end
	end

	-- area.ShowCursorLocation
	local function areaShowCursorLocation(self)
		if (not self:MouseInArea()) then
			self:HideCursorLocation()
			self:RemoveHighlight()
			if (GameTooltip:IsOwned(self)) then
				GameTooltip:Hide()
			end
			self.lastTooltipInfo = del(self.lastTooltipInfo)
			return
		end

		local x, y = GetCursorPosition()
		x, y = x / self:GetEffectiveScale(), y / self:GetEffectiveScale()
		if (self.lastX == x and self.lastY == y) then
			return
		end
		self.lastX, self.lastY = x, y

		local offset = -self.rightScroll:GetHorizontalScroll()
		local left = self.rightScroll:GetLeft()
		local cursorTime = ((x - left) + offset) / self.scale
		self.cursorTime = cursorTime

		local startSec = offset / self.scale
		local endSec = (offset + self.rightScroll:GetWidth() - 1) / self.scale

		local tooltipShown
		if (cursorTime < startSec or cursorTime > endSec) then
			self:HideCursorLocation()
			self:RemoveHighlight()
		else
			if (not self.cursorTick) then
				self.cursorTick = self:CreateSlidingIndicator()
				self.cursorTick:SetHeight(24)
				self.cursorTick:SetVertexColor(0, 1, 0)
				self.cursorTick:SetTexCoord(0, 1, 0.14, 0.9)
			end

			self.cursorTick:Show()
			self.cursorTick:SetPoint("BOTTOM", self.timeStart, "BOTTOM", (cursorTime - startSec) * self.scale, 0)

			if (cursorTime < 60) then
				self.cursorTick.text:SetFormattedText("%.3fs", cursorTime)
			else
				self.cursorTick.text:SetFormattedText("%dm %.3fs", cursorTime / 60, cursorTime % 60)
			end
			self.cursorTick.text:Show()

			-- Now highlight the spell row for where we are
			local left, top, width, height = self.lines[1]:GetRect()
			local row = floor((top - y) / height) + 2

			if (self.rowHighlight ~= row) then
				self:RemoveHighlight()
			end
			if (row >= 1 and row <= #self.lines) then
				local line = self.lines[row]
				line:LockHighlight()
				self.rowHighlight = row

				if (MouseIsOver(self)) then
					-- Now see if any spells are under the cursor, and show the info
					-- We cache what we show and compare this with last tooltip. Because we're updating
					-- so often there's a noticable lag when updating the tooltip at this frequency
					local temp = new()
					local cursorTimestamp = self.fight.combatStart + cursorTime
					for i,block in ipairs(line.blocks.list) do
						if (cursorTimestamp >= block.timeStart and cursorTimestamp <= block.timeEnd) then
							local name, rank, tex = GetSpellInfo(block.spellId)
							local r, g, b = block:GetVertexColor()
							tinsert(temp, new(name, tex, block.sourceName, r, g, b, block.mobName))
						end
					end

					if (#temp > 0) then
						local redraw
						if (self.lastTooltipInfo) then
							if (#self.lastTooltipInfo ~= #temp) then
								redraw = true
							else
								for i = 1,#temp do
									local t1 = temp[i]
									local t2 = self.lastTooltipInfo[i]
									if (t1[1] ~= t2[1] or t1[3] ~= t2[3]) then
										redraw = true
										break
									end
								end
							end
						else
							redraw = true
						end

						local showTime = self.fight.combatStart + cursorTime
						showTime = format("%s.%03d", date("%X", showTime), showTime * 1000 % 1000)

						tooltipShown = true
						if (redraw) then
							local c = 0
							for key in pairs(self.mobKeys) do
								c = c + 1
							end

							GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
							GameTooltip:SetText(showTime, 1, 1, 1)

							for i,t in ipairs(temp) do
								if (c > 1) then
									GameTooltip:AddDoubleLine(format("|cFFFFFF80%s|r on %s", t[1], t[7]), t[3], 0.5, 0.5, 0.5, t[4], t[5], t[6])
								else
									GameTooltip:AddDoubleLine(t[1], t[3], 1, 1, 0.5, t[4], t[5], t[6])
								end
								GameTooltip:AddTexture(t[2])
							end
							GameTooltip:Show()
						else
							-- Always update the cursor location time
							GameTooltipTextLeft1:SetText(showTime)
							GameTooltipTextLeft1:SetTextColor(1, 1, 1)
						end
					end

					del(self.lastTooltipInfo)
					self.lastTooltipInfo = temp
				end
			end
		end

		if (not tooltipShown) then
			GameTooltip:Hide()
			self.lastTooltipInfo = del(self.lastTooltipInfo)
		end
	end

	-- area.RemoveHighlight
	local function areaRemoveHighlight(self)
		if (self.rowHighlight) then
			self.lines[self.rowHighlight]:UnlockHighlight()
			self.rowHighlight = nil
		end
	end

	-- area.ShowScroll
	local function areaShowScroll(self)
		local offset = -self.rightScroll:GetHorizontalScroll()

		if (not self.timeStart) then
			self.timeStart = self:CreateSlidingIndicator()
			self.timeStart:SetPoint("BOTTOM", self.rightScroll, "TOPLEFT", 0, 0)
		end
		if (not self.timeEnd) then
			self.timeEnd = self:CreateSlidingIndicator()
			self.timeEnd:SetPoint("BOTTOM", self.rightScroll, "TOPRIGHT", -1, 0)
		end

		local width = self.rightScroll:GetWidth() - 1
		local startSec = offset / self.scale
		local endSec = (offset + width) / self.scale

		self.timeStart.text:SetFormattedText("%d", abs(startSec))
		self.timeEnd.text:SetFormattedText("%d", endSec)

		if (not self.ticks) then
			self.ticks = {}
		end

		local div = self.scale < 4 and 30 or self.scale < 8 and 15 or self.scale < 16 and 5 or 1
		local divText = self.scale < 4 and 60 or self.scale < 8 and 30 or self.scale < 16 and 15 or 5

		local startPoint = floor((startSec + div) / div) * div
		local i = 0
		for secs = startPoint,endSec,div do
			local short = secs % divText ~= 0
			local vshort = secs % 5 ~= 0
			local gap = short and 20 or 50

			local showOffset = (secs - startSec) * self.scale

			if (showOffset >= width - 10) then
				break
			elseif (showOffset > 10) then
				local nearEdge = min(showOffset - 10, width - 10 - showOffset)
				local fade
				if (nearEdge < gap) then
					fade = nearEdge / gap
				else
					fade = 1
				end

				i = i + 1
				local tick = self.ticks[i]
				if (not tick) then
					tick = self:CreateSlidingIndicator()
					self.ticks[i] = tick
				end

				tick:SetPoint("BOTTOM", self.timeStart, "BOTTOM", (secs - startSec) * self.scale, 0)
				tick:Show()

				if (vshort) then
					tick:SetVertexColor(0.5, 0.5, 0.5)
				else
					tick:SetVertexColor(1, 1, 1)
				end

				if (short) then
					tick:SetHeight(3)
					tick:SetTexCoord(0, 1, 0.5, 0.9)
					tick.text:Hide()
				else
					tick:SetHeight(8)
					tick:SetTexCoord(0, 1, 0.05, 0.5)
					tick.text:SetFormattedText("%d:%02d", secs / 60, secs % 60)

					tick.text:Show()
				end

				tick:SetAlpha(fade)
				tick.text:SetAlpha(fade)
			end
		end
		for j = i + 1,#self.ticks do
			local tick = self.ticks[j]
			tick:Hide()
			tick.text:Hide()
		end
	end

	-- area.BuildSpellNameList
	local function areaBuildSpellNameList(self)
		if (not self.mobKeys) then
			return
		end

		local fight = self.fight
		local data = fight.data

		local spellIcons = new()
		local spellIds = new()
		local lineColours = new()
		local activeClassesForCat = new()

		local temp = new()
		for mobKey in pairs(self.mobKeys) do
			local fightData = data[mobKey]

			for i = 1,#fightData do
				local segment = fightData[i]
				local timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId = module:GetStoredData(fight, segment)
				local name, rank, tex = GetSpellInfo(spellId)
				local spellInfo = Utopia.lookup.debuffs[name]
				if (self.mode == "category") then
					if (spellInfo) then
						name = spellInfo.key
						if (db.coloursExtreme) then
							local cat = Utopia.debuffs[name]
							local temp2 = activeClassesForCat[cat.key]
							if (not temp2) then
								temp2 = new()
								activeClassesForCat[cat.key] = temp2
							end
							temp2[spellInfo.class] = true
						end

						if (spellInfo.multiple) then
							for i,key in ipairs(spellInfo.multiple) do
								temp[key] = true

								if (db.coloursExtreme) then
									local temp2 = activeClassesForCat[key]
									if (not temp2) then
										temp2 = new()
										activeClassesForCat[key] = temp2
									end
									temp2[spellInfo.class] = true
								end
							end
						end
					end
				else
					spellIcons[name] = tex
					spellIds[name] = spellId
					local c = RAID_CLASS_COLORS[spellInfo.class]
					lineColours[name] = new(c.r, c.g, c.b)
				end

				temp[name] = true
			end
		end

		local toSort = new()
		for name in pairs(temp) do
			tinsert(toSort, name)
		end
		del(temp)
		sort(toSort)

		for i,name in ipairs(toSort) do
			local displayLine = self.nameToLineLookup[name]
			if (not displayLine) then
				self.shownLines = self.shownLines + 1
				displayLine = self.shownLines
				self.nameToLineLookup[name] = displayLine
			end

			-- Pre-create the known spell lines, so they appear in correct order later
			local line = self.lines[displayLine]
			if (not line) then
				line = self:CreateSpellLine(displayLine)
			end
			line.text:SetText(name)
			line:Show()

			if (self.mode == "category") then
				local cat = Utopia.debuffs[name]
				line.icon:SetTexture(cat.icon)
				line.text:SetTextColor(0.8, 0.8, 0.8)
			else
				line.icon:SetTexture(spellIcons[name])
			end
			line.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

			if (self.mode == "spell") then
				line.spellId = spellIds[name]
				line.text:SetTextColor(unpack(lineColours[name]))
			else
				if (db.coloursExtreme) then
					local cat = Utopia.debuffs[name]

					local activeClasses = activeClassesForCat[cat.key]
					local temp = new()
					for class in pairs(activeClasses) do
						tinsert(temp, class)
					end
					sort(temp)

					lineColours[name] = new()
					for i,class in pairs(temp) do
						local c = RAID_CLASS_COLORS[class]
						tinsert(lineColours[name], new(c.r, c.g, c.b))
					end
					del(temp)

					local str
					local lc = lineColours[name]
					if (#lc > 2) then
						str = ""
						for i = 1,#lc - 1 do
							local r1, g1, b1 = unpack(lc[i])
							local r2, g2, b2 = unpack(lc[i + 1])
	
							local part = name:sub(floor((i - 1) * (#name / (#lc - 1)) + 1), floor(i * (#name / (#lc - 1))))
							part = Gradient(part, r1, g1, b1, r2, g2, b2)
							str = str.. part
						end
					elseif (#lc == 2) then
						local r1, g1, b1 = unpack(lc[1])
						local r2, g2, b2 = unpack(lc[2])
						str = Gradient(name, r1, g1, b1, r2, g2, b2)
					else
						line.text:SetTextColor(unpack(lc[1]))
					end
					if (str) then
						line.text:SetText(str)
					end
				end
			end
		end

		--self.leftScrollChild:SetHeight(self.shownLines * 20)
		--self.rightScrollChild:SetHeight(self.shownLines * 20)

		del(toSort)
		del(spellIcons)
		del(spellIds)
		del(lineColours)
	end

	-- area.AddSpell
	local function areaAddSpell(self, timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey)
		local spellName, rank, tex = GetSpellInfo(spellId)
		local spellInfo = Utopia.lookup.debuffs[spellName]
		local name
		if (self.mode == "category") then
			name = spellInfo.key
		else
			name = spellName
		end

		self:AddSpellInfoToLine(timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey, name, spellInfo)

		if (self.mode == "category" and spellInfo.multiple) then
			for i,key in ipairs(spellInfo.multiple) do
				self:AddSpellInfoToLine(timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey, key, spellInfo)
			end
		end
	end

	-- area.AddSpellInfoToLine
	local function areaAddSpellInfoToLine(self, timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey, name, spellInfo)
		local xOffsetStart = (timeStart - self.fight.combatStart) * self.scale
		local xWidth = (timeEnd - timeStart) * self.scale
		if (xWidth < 0.05) then
			return
		end

		local displayLine = self.nameToLineLookup[name]
		if (not displayLine) then
			self.shownLines = self.shownLines + 1
			displayLine = self.shownLines
			self.nameToLineLookup[name] = displayLine
		end

		local line = self.lines[displayLine]
		if (not line) then
			line = self:CreateSpellLine(displayLine)
		end

		local tex = module:AquireAreaTexture(line.blocks, self.playerNames)
		tex:SetPoint("TOPLEFT", xOffsetStart, 0)
		tex:SetWidth(xWidth)

		local c = RAID_CLASS_COLORS[spellInfo.class]
		local r, g, b = c.r, c.g, c.b
		if (displayLine % 2 == 1) then
			r, g, b = max(0, r * 0.9), max(0, g * 0.9), max(0, b * 0.9)
		end
		tex:SetVertexColor(r, g, b)

		if (self.playerNames) then
			tex:SetOwnerText(sourceName, r, g, b)
		end

		tex.spellId		= spellId
		tex.realSpellId	= realSpellId
		tex.sourceName	= sourceName
		tex.timeStart	= timeStart
		tex.timeEnd		= timeEnd
		tex.mobName		= mobName
		tex.mobKey		= mobKey

		tinsert(line.blocks.list, tex)
	end

	do
		local function slOnMouseDown(self)
			local area = self:GetParent():GetParent():GetParent()
			area.rightScroll:OnMouseDown()
		end

		local function slOnMouseUp(self)
			local area = self:GetParent():GetParent():GetParent()
			area.rightScroll:OnMouseUp()
		end

		local function slOnEnter(self)
			if (self.spellId) then
				local area = self:GetParent():GetParent():GetParent()
				local rs = area.rightScroll
				if (not rs.dragging and not rs.speedX and not rs.swooshTarget and not rs.targetTime) then
					GameTooltip:SetOwner(self, "ANCHOR_TOP")
					GameTooltip:SetHyperlink("spell:"..self.spellId)
				end
			end
		end

		local function slOnLeave(self)
			GameTooltip:Hide()
		end

		local function slOnShow(self)
			-- Also, we need to hide this when we hide the line else the ScrollFrame
			-- will keep the max growth height we've had even with hidden lines
			self.blocks:Show()
		end

		local function slOnHide(self)
			local list = self.blocks.list
			for i,tex in pairs(list) do
				module:ReleaseAreaTexture(tex)
			end
			del(list)
			self.blocks.list = new()
			self.blocks:Hide()
		end

		-- CreateSpellLine
		function areaCreateSpellLine(self, index)
			local l = CreateFrame("Button", nil, self.leftScrollChild)
			self.lines[index] = l
			l:EnableMouse(true)

			-- NOTE: Have to use absolute anchoring here only to the ScrollChild because of ScrollFrame limitations
			l:SetPoint("TOPLEFT", 0, -(index - 1) * 20)
			l:SetPoint("BOTTOMRIGHT", self.leftScrollChild, "TOPRIGHT", 0, -index * 20)

			if (index % 2 == 0) then
				l.bg = l:CreateTexture(nil, "BACKGROUND")
				l.bg:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
				l.bg:SetVertexColor(0.3, 0.3, 0.8, 0.3)
				l.bg:SetAllPoints()
			end

			l.icon = l:CreateTexture(nil, "OVERLAY")
			l.icon:SetWidth(18)
			l.icon:SetHeight(18)
			l.icon:SetPoint("TOPLEFT", 1, -1)

			l.text = l:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			l.text:SetHeight(20)
			l.text:SetWidth(300)
			l.text:SetPoint("LEFT", l.icon, "RIGHT", 2, 0)
			l.text:SetJustifyH("LEFT")

			l:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
			l.highlight = l:GetHighlightTexture()
			l.highlight:SetBlendMode("ADD")
			l.highlight:SetTexCoord(0.25, 0.75, 0.25, 0.75)
			l.highlight:SetAlpha(0.3)
			l.highlight:SetAllPoints(true)

			l.blocks = CreateFrame("Frame", nil, self.rightScrollChild)
			l.blocks.list = {}

			-- NOTE: Have to use absolute anchoring here only to the ScrollChild because of ScrollFrame limitations
			l.blocks:SetPoint("TOPLEFT", 0, -(index - 1) * 20)
			l.blocks:SetPoint("BOTTOMRIGHT", self.rightScrollChild, "TOPRIGHT", 0, -index * 20)

			l:SetScript("OnMouseDown", slOnMouseDown)
			l:SetScript("OnMouseUp", slOnMouseUp)
			l:SetScript("OnEnter", slOnEnter)
			l:SetScript("OnLeave", slOnLeave)
			l:SetScript("OnHide", slOnHide)
			l:SetScript("OnShow", slOnShow)

			return l		
		end
	end


	local areaOnMouseDown, areaOnMouseUp
	do
		-- Dragging Functions
		local function draggingOnUpdate(self, elapsed)
			local x, y = GetCursorPosition()
			x, y = x / self:GetEffectiveScale(), y / self:GetEffectiveScale()

			local curX, curY = -self:GetHorizontalScroll(), self:GetVerticalScroll()
	    	local newX = min(self:GetHorizontalScrollRange(), max(0, curX - x + self.dragX))
	    	local newY = min(self:GetVerticalScrollRange(), max(0, curY + y - self.dragY))

			self:SetHorizontalScroll(-newX)
			self:SetVerticalScroll(newY)
			self:GetParent():ShowCursorLocation()

			local scale = self:GetParent().scale
			self.speedX = (newX - curX)
			self.speedY = newY - curY
			self.dragX, self.dragY = x, y
		end

		-- driftingOnUpdate
		local function driftingOnUpdate(self, elapsed)
			local curX, curY = -self:GetHorizontalScroll(), self:GetVerticalScroll()
			local area = self:GetParent()

			local scale = area.scale
			self.speedX = self.speedX * (1 - elapsed * 10)
			self.speedY = self.speedY * (1 - elapsed * 10)

	    	local newX = min(self:GetHorizontalScrollRange(), max(0, curX + self.speedX))
	    	local newY = min(self:GetVerticalScrollRange(), max(0, curY + self.speedY))

			self:SetHorizontalScroll(-newX)
			self:SetVerticalScroll(newY)
			self:GetParent():ShowCursorLocation()

			if (abs(self.speedX) < 0.1 and abs(self.speedY) < 0.1) then
				self.speedX, self.speedY = nil, nil
				if (area:MouseInArea()) then
					self:SetScript("OnUpdate", areaRightScrollOnUpdate)
				else
					self:SetScript("OnUpdate", nil)
				end
				return
			end
		end

		-- area.MouseInArea
		function areaMouseInArea(self)
			if (MouseIsOver(self)) then
				return true
			end
			local dpsFrame = self:GetParent().dpsFrame
			if (dpsFrame) then
				return MouseIsOver(dpsFrame)
			end
		end

		-- OnMouseDown
		function areaOnMouseDown(self)
			self.dragging = true
			self.dragX, self.dragY = GetCursorPosition()
			self.dragX, self.dragY = self.dragX / self:GetEffectiveScale(), self.dragY / self:GetEffectiveScale()
			self:SetScript("OnUpdate", draggingOnUpdate)
		end

		-- OnMouseUp
		function areaOnMouseUp(self)
			self.dragging = nil
			if (max(abs(self.speedX or 0), abs(self.speedY or 0)) >= 0.1) then
				self:SetScript("OnUpdate", driftingOnUpdate)
			else
				self.speedX, self.speedY = nil, nil
				if (self:GetParent():MouseInArea()) then
					self:SetScript("OnUpdate", areaRightScrollOnUpdate)
				else
					self:SetScript("OnUpdate", nil)
				end
			end
		end
	end

	-- Zoom
	local function areaZoom(self, zoomIn)
		-- Change the scale
		local oldScale = self.scale
		local diff = floor((self.scale + 10) / 10)			-- 1-10 steps 1, 11-20 steps 2 etc.
		if (zoomIn) then
			self.scale = min(64, self.scale + diff)
		else
			self.scale = max(1, self.scale - diff)
		end
		if (self.scale == oldScale) then
			return
		end

		-- Work out where our cursor is atm, we'll want it in the same position after zooming
		local offset = -self.rightScroll:GetHorizontalScroll()
		local x = GetCursorPosition() / self:GetEffectiveScale()
		local cursorOffset = x - self.rightScroll:GetLeft()
		local cursorTime = (cursorOffset + offset) / oldScale
		local startSec = offset / oldScale
		if (cursorTime < startSec) then
			cursorTime = startSec	-- Mouse is not over the ScrollFrame (probably using the controls for zooming)
		end

		-- Redraw
		self:Init()
		--self.dataIndex = del(self.dataIndex)
		--self.dataIndex = new()
		--self:ResetLines()
		self:Populate()
		self:ShowScale()

		-- Now set the new offset so the cursor is where it was before zooming
		local newOffset = (cursorTime * self.scale) - cursorOffset
		if (newOffset < 0) then
			newOffset = 0
		end
		self.rightScroll:SetHorizontalScroll(-newOffset)
	end

	-- OnMouseWheel
	local function areaOnMouseWheel(self, delta)
		local area = self:GetParent()
		if (delta < 0) then
			area:Zoom()
		else
			area:Zoom(true)
		end
	end

	-- ShowScale
	-- Internally, a scale of 1 means 1 pixel to 1 second
	-- That's a bit small tho, 6 pixels to 1 second is a good visual point, so we'll call that the 100% mark
	local function areaShowScale(self)
		self.scaleFrame.text:SetFormattedText(L["Zoom: %d%%"], 100 / 6 * self.scale)
	end

	-- ScaleUpOnClick
	local function areaScaleUpOnClick(self)
		local area = self:GetParent():GetParent().area
		area:Zoom(true)
	end

	-- ScaleDownOnClick
	local function areaScaleDownOnClick(self)
		local area = self:GetParent():GetParent().area
		area:Zoom()
	end

	-- area.rightScroll:OnSizeChanged
	local function areaOnSizeChanged(self, x, y)
		self:GetParent():ShowScroll()
	end

	-- area:OnVerticalScroll
	local function areaOnVerticalScroll(self, offset)
		local rightScrollbar = getglobal(self:GetName().."ScrollBar")
		self:GetParent().leftScroll:SetVerticalScroll(rightScrollbar:GetValue())
		self:GetParent():ShowScroll()
	end

	-- area.rightScroll:OnHorizontalScroll
	local function areaOnHorizontalScroll(self, offset)
		local scrollbar = getglobal(self:GetName().."ScrollBar2")
		if (scrollbar) then
			scrollbar:SetValue(offset)
			local min, max = scrollbar:GetMinMaxValues()
			if ( offset <= 0 ) then
				getglobal(scrollbar:GetName().."ScrollUpButton"):Disable()
			else
				getglobal(scrollbar:GetName().."ScrollUpButton"):Enable()
			end
			if (offset >= max) then
				getglobal(scrollbar:GetName().."ScrollDownButton"):Disable()
			else
				getglobal(scrollbar:GetName().."ScrollDownButton"):Enable()
			end
		end

		local area = self:GetParent()
		local frame = area:GetParent()
		area:ShowScroll()
		if (frame.dpsFrame and (frame.showDPS or frame.showDPSBoss)) then
			frame.dpsFrame.scroll:SetHorizontalScroll(self:GetHorizontalScroll())
		end
	end

	-- CreateWorkArea
	function module:CreateWorkArea(parent)
		local a = CreateFrame("Frame", nil, parent)
		-- info.record["0xF13000808A015559"]["0x010000000008BC20"]["Improved Scorch"]
		a.lines = {}
		a.mode = "category"
		a.scale = 6

		local leftScrollName = self:GenericSuffixName(parent, "ScrollFrame")
		local rightScrollName = self:GenericSuffixName(parent, "ScrollFrame")

		a.leftScroll = CreateFrame("ScrollFrame", leftScrollName, a)
		local ls = a.leftScroll
		a.leftScrollChild = CreateFrame("Frame", nil, ls)

		a.bar = self:CreateVerticalBar(a.leftScroll, "RIGHT", true)

		a.rightScroll = CreateFrame("ScrollFrame", rightScrollName, a, "UIPanelScrollFrameTemplate")
		local rs = a.rightScroll
		a.rightScrollChild = CreateFrame("Frame", nil, rs)

		ls:SetScrollChild(a.leftScrollChild)
		rs:SetScrollChild(a.rightScrollChild)

		ls:SetPoint("TOPLEFT")
		ls:SetPoint("BOTTOMRIGHT", a, "BOTTOMLEFT", 150, 0)
		rs:SetPoint("TOPLEFT", ls, "TOPRIGHT", 10, 0)
		rs:SetPoint("BOTTOMRIGHT")

		a.leftScrollChild:SetWidth(200)
		a.leftScrollChild:SetHeight(300)
		a.rightScrollChild:SetWidth(500)
		a.rightScrollChild:SetHeight(50)

		rs:HookScript("OnVerticalScroll", areaOnVerticalScroll)
		rs:SetScript("OnHorizontalScroll", areaOnHorizontalScroll)

		-- Header
		local htl = a:CreateTexture(nil, "BORDER")
		a.headerTopL = htl
		htl:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		htl:SetTexCoord(0, 0.25, 0, 1)
		htl:SetPoint("BOTTOMLEFT", a, "TOPLEFT", 120, -24)
		htl:SetWidth(64)
		htl:SetHeight(64)

		local htr = a:CreateTexture(nil, "BORDER")
		a.headerTopR = htr
		htr:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		htr:SetTexCoord(0.75, 1, 0, 1)
		htr:SetPoint("BOTTOMRIGHT", a, "TOPRIGHT", 30, -24)
		htr:SetWidth(64)
		htr:SetHeight(64)

		local hmid = a:CreateTexture(nil, "BORDER")
		a.headerMid = hmid
		hmid:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		hmid:SetTexCoord(0.0625, 0.9375, 0, 1)
		hmid:SetPoint("TOPLEFT", htl, "TOPRIGHT")
		hmid:SetPoint("BOTTOMRIGHT", htr, "BOTTOMLEFT")

		rs:EnableMouseWheel(true)
		rs:EnableMouse(true)
		rs:SetScript("OnMouseDown", areaOnMouseDown)
		rs:SetScript("OnMouseUp", areaOnMouseUp)
		rs:SetScript("OnSizeChanged", areaOnSizeChanged)
		rs:SetScript("OnEnter", areaRightScrollOnEnter)
		rs:SetScript("OnLeave", areaRightScrollOnLeave)
		rs:SetScript("OnMouseWheel", areaOnMouseWheel)

		rs.OnEnter = areaRightScrollOnEnter
		rs.OnLeave = areaRightScrollOnLeave
		rs.OnMouseDown = areaOnMouseDown
		rs.OnMouseUp = areaOnMouseUp

		-- Scale controls
		local sc = CreateFrame("Frame", nil, parent)
		a.scaleFrame = sc
		sc:SetPoint("BOTTOMLEFT", a, "TOPLEFT", 0, 10)
		sc:SetPoint("TOPRIGHT", htl, "TOPLEFT", -5, 0)

		sc.upButton = module:CreateArrowButton(sc, "up")
		sc.downButton = module:CreateArrowButton(sc, "down")
		sc.upButton:SetScript("OnClick", areaScaleUpOnClick)
		sc.downButton:SetScript("OnClick", areaScaleDownOnClick)

		sc.upButton:SetPoint("TOPRIGHT")
		sc.downButton:SetPoint("TOP", sc.upButton, "BOTTOM")
		sc.text = sc:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		sc.text:SetPoint("RIGHT", sc.upButton, "BOTTOMLEFT", -4, 0)

		-- Functions
		a.SelectMobKey			= areaSelectMobKey
		a.OnClose				= areaOnClose
		a.ResetLines			= areaResetLines
		a.ClearDisplay			= areaClearDisplay
		a.SetMode				= areaSetMode
		a.Init					= areaInit
		a.BuildSpellNameList	= areaBuildSpellNameList
		a.Populate				= areaPopulate
		a.PopulateAuras			= areaPopulateAuras
		a.PopulateDeaths		= areaPopulateDeaths
		a.CleanupDeaths			= areaCleanupDeaths
		a.ShowScroll			= areaShowScroll
		a.ShowScale				= areaShowScale
		a.ShowCursorLocation	= areaShowCursorLocation
		a.HideCursorLocation	= areaHideCursorLocation
		a.RemoveHighlight		= areaRemoveHighlight
		a.AddSpell				= areaAddSpell
		a.AddSpellInfoToLine	= areaAddSpellInfoToLine
		a.CreateSpellLine		= areaCreateSpellLine
		a.CreateSlidingIndicator = areaCreateSlidingIndicator
		a.Zoom					= areaZoom
		a.ShowDPS				= areaShowDPS
		a.CreateDeathMarker		= areaCreateDeathMarker
		a.AquireDeathMarker		= areaAquireDeathMarker
		a.ReleaseDeathMarker	= areaReleaseDeathMarker
		a.MouseInArea			= areaMouseInArea

		return a
	end
end

do
	-- dps.SetFight
	local function dpsSetFight(self, fight)
		self.fight = fight
		self:CalcDPS()
		self:DrawDPS()
	end

	-- dps.OnHide
	local function dpsOnHide(self)
		self.fight = nil
		self.dmg = deepDel(self.dmg)
		self.dpsMode = "raid"
		self:Cleanup()
	end

	-- dps.Cleanup
	local function dpsCleanup(self)
		if (self.texList) then
			for i,tex in ipairs(self.texList) do
				self:ReleaseDPSBlock(tex)
			end
			self.texList = del(self.texList)
		end
	end

	local blocks = {}
	-- dps.CreateDPSBlock
	local function dpsCreateDPSBlock(self, parent)
		local tex = parent:CreateTexture(nil, "OVERLAY")
		tex:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		return tex
	end

	-- dps.AquireDPSBlock
	local function dpsAquireDPSBlock(self, parent)
		local tex = tremove(blocks) or self:CreateDPSBlock(parent)
		tex:SetParent(parent)
		tex:Show()
		return tex
	end

	-- dps.ReleaseDPSBlock
	local function dpsReleaseDPSBlock(self, tex)
		tex:ClearAllPoints()
		tex:SetParent(UIParent)
		tex:Hide()
		tex:SetHeight(0)
		tex.amount = nil
		tinsert(blocks, tex)
	end

	-- dps.DrawDPS
	local function dpsDrawDPS(self)
		self:Cleanup()
		self.texList = new()

		local parent = self:GetParent()
		if (parent.showDPS and self.fight.dps) then
			self:DrawDPSPart("raid")
		end
		if (parent.showDPSBoss and self.fight.bossdps) then
			self:DrawDPSPart("boss")
		end

		local area = self:GetParent().area
		self.scroll:SetHorizontalScroll(area.rightScroll:GetHorizontalScroll())
	end

	-- dps.DrawDPS
	local function dpsDrawDPSPart(self, what)
		local dps = self.dmg and self.dmg[what]
		if (not self.dmg) then
			return
		end

		local endSec = floor(self.fight.combatEnd - self.fight.combatStart + 1)
		local area = self:GetParent().area
		local oneSecWidth = area.scale
		local height = self.height - 1

		for i = 1,endSec do
			local amount = dps.amounts[i]
			if (amount and amount > 0) then
				local tex = self:AquireDPSBlock(self.scrollChild)
				tinsert(self.texList, tex)

				tex:SetWidth(oneSecWidth)
				tex:SetHeight(amount / dps.max * height)
				tex:SetPoint("BOTTOMLEFT", oneSecWidth * (i - 1), 0)
				tex.amount = amount

				if (what == "boss") then
					tex:SetVertexColor(1, 0, 0)
				else
					tex:SetVertexColor(0, 1, 0)
				end
			end
		end
	end

	-- dps.CalcDPS
	local function dpsCalcDPS(self)
		self.dmg = deepDel(self.dmg)
		self.dmg = new()
		self.dmg.raid = new()

		self.dmg.raid.amounts, self.dmg.raid.max = self:CalcDPSPart(self.fight.dps)

		if (self.fight.bossdps) then
			self.dmg.boss = new()
			self.dmg.boss.amounts, self.dmg.boss.max = self:CalcDPSPart(self.fight.bossdps)
		end
	end

	-- dps.CalcDPSPart
	local function dpsCalcDPSPart(self, dps)
		local amounts = new()
		local maxDamage = 0

		local area = self:GetParent().area

		local preSmoothOffset, endSmoothOffset, slider
		local endSec = floor(self.fight.combatEnd - self.fight.combatStart + 1)
		if (db.smoothDPS > 1) then
			preSmoothOffset = floor(db.smoothDPS / 2)
			endSmoothOffset = db.smoothDPS - preSmoothOffset - 1
		end

		for mobKey,data in pairs(dps) do
			if (area.mobKeys[mobKey]) then
				local temp = new(strsplit(",", data))

				for i,dmg in ipairs(temp) do
					local amount
					if (db.smoothDPS > 1) then
						local average, count = 0, 0
						for j = i - preSmoothOffset, i + endSmoothOffset do
							local val = tonumber(temp[j])
							if (val) then
								count = count + 1
								average = average + val
							end
						end
						amount = average / count
					else
						amount = tonumber(dmg)
					end

					amounts[i] = max(amounts[i] or 0, amount)
					maxDamage = max(maxDamage, amount)
				end

				del(temp)
			end
		end

		return amounts, maxDamage
	end

	local function dpsOnMouseDown(self)
		local area = self:GetParent().area
		area.rightScroll:OnMouseDown()
	end

	local function dpsOnMouseUp(self)
		local area = self:GetParent().area
		area.rightScroll:OnMouseUp()
	end

	local function dpsOnEnter(self)
		local area = self:GetParent().area
		area.rightScroll:OnEnter()
	end

	local function dpsOnLeave(self)
		local area = self:GetParent().area
		area.rightScroll:OnLeave()
	end

	local function dpsGetSmoothValue(self)
		return db.smoothDPS
	end

	local function dpsOnSmoothChanged(self, value)
		db.smoothDPS = value
		self:CalcDPS()
		self:DrawDPS(true)
	end

	-- CreateDPSFrame
	function module:CreateDPSFrame(parent)
		local dps = CreateFrame("Frame", nil, parent)

		dps:EnableMouse(true)
		dps:SetScript("OnMouseDown", dpsOnMouseDown)
		dps:SetScript("OnMouseUp", dpsOnMouseUp)
		dps:SetScript("OnEnter", dpsOnEnter)
		dps:SetScript("OnLeave", dpsOnLeave)

		dps.smoothSlider = module:Slider(dps, "Smooth DPS", dpsGetSmoothValue, dpsOnSmoothChanged, 1, 20, 1, "BOTTOMLEFT", 10, 15)
		dps.smoothSlider.tooltip = L["DPS Smoothing will average out the dmg over the given number of seconds to provide a more useful visual representation of the damage output"]

		local scrollName = self:GenericSuffixName(parent, "ScrollFrame")
		local scroll = CreateFrame("ScrollFrame", scrollName, dps)
		dps.scroll = scroll
		dps.height = 100
		dps.dpsMode = "raid"

		local scrollChild = CreateFrame("Frame", nil, scroll)
		dps.scrollChild = scrollChild
		scroll:SetScrollChild(scrollChild)

		scrollChild:SetWidth(400)
		scrollChild:SetHeight(dps.height - 1)

		local offset = parent.area.rightScroll:GetLeft() - parent.area.leftScroll:GetLeft()

		scroll:SetPoint("TOPLEFT", offset, 0)
		scroll:SetPoint("BOTTOMRIGHT")

	    dps.SetFight		= dpsSetFight
	    dps.OnHide			= dpsOnHide
	    dps.Cleanup			= dpsCleanup
	    dps.DrawDPS			= dpsDrawDPS
	    dps.DrawDPSPart		= dpsDrawDPSPart
	    dps.CalcDPS			= dpsCalcDPS
	    dps.CalcDPSPart		= dpsCalcDPSPart
		dps.CreateDPSBlock	= dpsCreateDPSBlock
		dps.AquireDPSBlock	= dpsAquireDPSBlock
		dps.ReleaseDPSBlock	= dpsReleaseDPSBlock

		return dps
	end
end

do
	local list = {}
	local listNames = {}

	-- AquireAreaTexture
	function module:AquireAreaTexture(frame, names)
		local tex
		if (names) then
			tex = tremove(listNames) or tremove(list) or self:CreateAreaTexture(frame)
		else
			tex = tremove(list) or tremove(listNames) or self:CreateAreaTexture(frame)
		end
		tex:SetParent(frame)
		if (tex.ownerText) then
			tex.ownerText:SetParent(frame)
		end
		tex:Show()
		return tex
	end

	-- ReleaseAreaTexture
	function module:ReleaseAreaTexture(tex)
		tex:ClearAllPoints()
		tex:Hide()
		tex:SetParent(UIParent)
		tex.spellId		= nil
		tex.realSpellId	= nil
		tex.sourceName	= nil
		tex.timeStart	= nil
		tex.timeEnd		= nil

		if (tex.ownerText) then
			tex.ownerText:Hide()
			tinsert(listNames, tex)
		else
			tinsert(list, tex)
		end
	end

	-- SetOwnerText
	local function setOwnerText(self, name, r, g, b)
		-- For some reason this will bump the position of the textures over a little
		-- Tried the alternatively anchoring the text blocks to the ScrollChild instead of the parent texture
		-- shown in the commented-out code with exact same results

		if (not self.ownerText) then
			self.ownerText = self:GetParent():CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			--self.ownerText:SetHeight(20)
			self.ownerText:SetPoint("LEFT", self, "LEFT")
		else
			self.ownerText:Show()
		end
		self.ownerText:SetText(name)
		self.ownerText:SetTextColor(r or 0, g or 0, b or 0)

		--local point, _, rel, x, y = self:GetPoint(1)
		--self.ownerText:SetPoint(point, self:GetParent(), rel, x, y)
	end

	-- CreateAreaTexture
	function module:CreateAreaTexture(frame)
		local tex = frame:CreateTexture(nil, "BACKGROUND")
		tex:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		tex:SetVertexColor(0, 0.8, 0)
		tex:SetHeight(20)
		tex.SetOwnerText = setOwnerText
		return tex
	end
end

-- DescribeDate
function module:DescribeDate(describe)
	local today = date("%x", time())
	local yesterday = date("%x", time() - 24*60*60)
	local week = time() - 7*24*60*60
	local when = date("%x", describe)
	local age
	if (when == today) then
		when = L["Today"]
		age = 1
	elseif (when == yesterday) then
		when = L["Yesterday"]
		age = 2
	elseif (describe > week) then
		when = date("%A", describe)
		age = 3
	else
		when = date("%X", describe)
		age = 4
	end
	return when, age
end

do
	-- mainlist.OnShow
	local function mainlistOnShow(self)
		self:Populate()
		self.list:Click(1)
	end

	-- mainlist.OnHide
	local function mainlistOnHide(self)
		self.list:Clear()
		self.list2:Clear()
	end

	-- mainlist.ShowStats
	local function mainlistShowStats(self)
		self.stats:SetFormattedText(L["History Size: %d"], db.history and #db.history or 0)
	end

	-- mainlist.ShowOptions
	local function mainlistShowOptions(self)
		local sel = new(self.list2:GetSelected())
		if (next(sel)) then
			self.delete:Show()
			self.open:Show()
		else
			self.delete:Hide()
			self.open:Hide()
		end
		del(sel)
	end

	-- mainlist.DeleteFights
	local function mainlistViewFights(self)
		local mainlist = self:GetParent()
		local list2 = mainlist.list2

		local sel = new(list2:GetSelected())
		for i,index in ipairs(sel) do
			local entry = list2.items[index]
			module:OpenView(entry[2])
		end
	end

	-- mainlist.DeleteFights
	local function mainlistDeleteFights(self)
		local mainlist = self:GetParent()
		local list2 = mainlist.list2

		local sel = new(list2:GetSelected())
		local delete = new()

		for i,index in ipairs(sel) do
			local key = list2.items[index][2]
--@debug@
			local info = db.history[key]
			local when = module:DescribeDate(info.combatStart)
			local display = format("%s: %s", info.player, when)
			assert(info, "info")
            assert(display == mainlist.key, "display == mainlist.key")
--@no-debug@
			tinsert(delete, key)
		end

		sort(delete)
		for i = #delete,1,-1 do
			deepDel(tremove(db.history, delete[i]))
		end

		del(sel)
		del(delete)

		mainlist:Populate()
		mainlist.list:Click(1)
	end

	-- list.OnClick
	local function mainlistOnClick(self, key)
		local parent = self:GetParent()
		parent:Populate2(key)
	end

	-- list2.OnClick
	local function mainlistOnClick2(self, key)
		local parent = self:GetParent()
		parent:ShowOptions()
	end

	-- list2.OnDoubleClick
	local function mainlistOnDoubleClick2(self, key)
		module:OpenView(key)
	end

	local ageColours = {"|cFF90FF90", "|cFFFFFF80", "|cFFFFA020", "|cFFFF4040"}

	-- Populate
	local function mainlistPopulate(self)
		if (db.history) then
			local temp = new()
			local listItems = new()

			for i = #db.history,1,-1 do
				local info = db.history[i]
				local when, age = module:DescribeDate(info.combatStart)
				local colour = ageColours[age]

				local key = format("%s: %s", info.player, when)
				if (key ~= last) then
					if (not temp[key]) then
						temp[key] = true
						last = key
						tinsert(listItems, new(format("%s: %s%s", info.player, colour, when), key))
					end
				end
			end
			del(temp)

			self.list:SetItems(listItems)
		end
		self:ShowOptions()
		self:ShowStats()
	end

	-- Populate2
	local function mainlistPopulate2(self, key)
		if (not db.history) then
			return
		end

		local listItems = new()
		self.key = key or self.key

		for i = #db.history,1,-1 do
			local info = db.history[i]
			local when = module:DescribeDate(info.combatStart)

			local display = format("%s: %s", info.player, when)
			if (display == self.key) then
				-- Pick an appropriate mob name to display
				local mobname
				if (info.boss) then
					mobname = info.boss
				else
					local _
					_, mobname = next(info.mobs)
					for j,name in pairs(info.mobs) do
						if (bossid[j]) then
							mobname = name			-- But use a boss' name if there's one
							break
						end
					end
				end

				local dur = info.combatEnd - info.combatStart
				if (dur > 60) then
					dur = format(L["%.1f mins"], dur / 60)
				else
					dur = format(L["%d secs"], dur)
				end

				tinsert(listItems, new(format("%s - %s - %s", date("%H:%M", info.combatStart), dur, mobname or ""), i))
			end
		end

		self.list2:SetItems(listItems)
		self:ShowOptions()
	end

	-- CreateMainList
	function module:CreateMainList()
		local frame = CreateFrame("Frame", "UtopiaUpTimeList", UIParent)
		frame:Hide()
		frame:SetWidth(512)
		frame:SetHeight(512)
		frame:SetScale(0.8)

		self:ApplyBackground(frame)
		self:MovableFrame(frame)

		frame:SetScript("OnHide", mainlistOnHide)
		frame:SetScript("OnShow", mainlistOnShow)

		frame.title:SetText(L["Fight Listing"])

		frame.list = self:CreateListFrame(frame, true)
		frame.list:SetPoint("TOPLEFT", 24, -82)
		frame.list:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", 200, 42)
		frame.list.OnClick = mainlistOnClick

		frame.bar = self:CreateVerticalBar(frame.list, "RIGHT")

		frame.list2 = self:CreateListFrame(frame, true, true)
		frame.list2:SetPoint("TOPLEFT", frame.list.verticalBar, "TOPRIGHT", 0, 0)
		frame.list2:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -40, 42)
		frame.list2.OnClick = mainlistOnClick2
		frame.list2.OnDoubleClick = mainlistOnDoubleClick2
		frame.list2.OnSelectMultiple = mainlistOnClick2

		frame.Populate = mainlistPopulate
		frame.Populate2 = mainlistPopulate2
		frame.ShowOptions = mainlistShowOptions
		frame.ShowStats = mainlistShowStats

		frame.delete = self:Button(frame, DELETE, mainlistDeleteFights, "BOTTOMLEFT", 15, 15)
		frame.open = self:Button(frame, L["Open"], mainlistViewFights, "BOTTOMRIGHT", -15, 15)

		frame.stats = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		frame.stats:SetPoint("TOPLEFT", 70, -25)
		frame.stats:SetPoint("BOTTOMRIGHT", frame.list2, "TOPRIGHT", 20, 5)
		frame.stats:SetJustifyH("RIGHT")

		return frame
	end
end

-- Toggle
function module:Toggle()
	if (not self.listFrame) then
		self.listFrame = self:CreateMainList()
	end
	if (self.listFrame:IsShown()) then
		self.listFrame:Hide()
	else
		self.listFrame:Show()
	end
end

-- OnInitialize
function module:OnInitialize()
	local defaults = { 
		profile = {
			bossesOnly = true,
			notifications = true,
			minimumDuration = 30,
			keep = 25,
			coloursExtreme = true,
			smoothDPS = 5,
		} 
	}

	-- If we're loading modular, then we'll store our DB associated with the module instead of Utopia proper
	-- Mostly, so I can easily check what the memory usage of this module's data will be like
	if (IsAddOnLoaded("Utopia_UpTime")) then
		self.db = LibStub("AceDB-3.0"):New("UtopiaUpTimeDB", defaults, "Default")
		self:MoveDataFromUtopiaDBtoLocalDB()
	else
		self.db = Utopia.db:RegisterNamespace("UpTime", defaults)
	end

	db = self.db.profile
end

local svList = {"history", "bossesOnly", "notifications", "minimumDuration", "keep"}

-- MoveDataFromUtopiaDBtoLocalDB
function module:MoveDataFromUtopiaDBtoLocalDB()
	if (Utopia.db.sv.namespaces.UpTime) then
		-- We moved from embeded to modular, move the settings over. (can't go the other way tho)
		local profiles = Utopia.db.sv.namespaces.UpTime.profiles
		local oldIntegratedDB = profiles and profiles.Default
		if (oldIntegratedDB and oldIntegratedDB.history and not self.db.profile.history) then
			for i,key in pairs(svList) do
				self.db.profile[key] = oldIntegratedDB[key]
				oldIntegratedDB[key] = nil
			end
		end
	end
end

-- MoveDataFromLocalDBtoUtopiaDB
function module:MoveDataFromLocalDBtoUtopiaDB()
	local ns = Utopia.db.sv.namespaces
	if (ns) then
		if (not ns.UpTime) then
			ns.UpTime = {profiles = {Default = {}}}
		end

		local newDB = ns.UpTime.profiles.Default

		for i,key in pairs(svList) do
			newDB[key] = self.db.profile[key]
			self.db.profile[key] = nil
		end
	end
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
end

-- DXEStartBossEncounter
function module:DXEStartBossEncounter()
	self.activeBossEncounter = self.currentBossEncounter
end

-- DXEStopBossEncounter
function module:DXEStopBossEncounter()
	self.activeBossEncounter = nil
end

-- DXESetActiveEncounter
function module:DXESetActiveEncounter(e, ce)
	self.currentBossEncounter = ce and ce.name or nil
end

function module:BigWigs_TargetSeen(mobname, unit, module)
	Xmodule = module
d("BigWigs_TargetSeen(%s, %s, %s)", tostring(mobname), tostring(unit), tostring(module))
	self.currentBossEncounter = module.name or module.bossname
end

-- OnEnable
function module:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_LEAVING_WORLD")
	self:RawHook("SetItemRef", true)

	if (DXE) then
		DXE:RegisterCallback("StartEncounter", self.DXEStartBossEncounter, self)
		DXE:RegisterCallback("StopEncounter", self.DXEStopBossEncounter, self)
		DXE:RegisterCallback("SetActiveEncounter", self.DXESetActiveEncounter, self)
	end
	if (BigWigs) then
		self:Hook(BigWigs, "BigWigs_TargetSeen", "BigWigs_TargetSeen")
	end

	if (InCombatLockdown()) then
		self:StartFighting()
	end

	FRIEND_FLAGS = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
	ENEMY_FLAGS = COMBATLOG_OBJECT_REACTION_HOSTILE
	EXLCUDE_FLAGS = COMBATLOG_OBJECT_TYPE_OBJECT + COMBATLOG_OBJECT_TYPE_PLAYER

	db = self.db.profile
end

-- OnDisable
function module:OnDisable()
	self.record, self.linear, self.combatStart = nil, nil, nil
end
